# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import json
import tempfile
from decimal import Decimal
from typing import Any, Union

import numpy as np
import pytest
from numpy.typing import NDArray
from PIL import Image

import streaming.base.format.json.encodings as jsonEnc
import streaming.base.format.mds.encodings as mdsEnc
import streaming.base.format.xsv.encodings as xsvEnc


class TestMDSEncodings:

    @pytest.mark.parametrize('data', [b'5', b'\x00\x00'])
    def test_byte_encode_decode(self, data: bytes):
        byte_enc = mdsEnc.Bytes()
        assert byte_enc.size is None
        output = byte_enc.encode(data)
        assert output == data
        output = byte_enc.decode(data)
        assert output == data

    @pytest.mark.parametrize('data', ['9', 25])
    def test_byte_encode_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            byte_enc = mdsEnc.Bytes()
            _ = byte_enc.encode(data)

    @pytest.mark.parametrize(('data', 'encode_data'),
                             [('99', b'99'), ('streaming dataset', b'streaming dataset')])
    def test_str_encode_decode(self, data: str, encode_data: bytes):
        str_enc = mdsEnc.Str()
        assert str_enc.size is None

        # Test encode
        enc_data = str_enc.encode(data)
        assert isinstance(enc_data, bytes)
        assert enc_data == encode_data

        # Test decode
        dec_data = str_enc.decode(encode_data)
        assert isinstance(dec_data, str)
        assert dec_data == data

    @pytest.mark.parametrize('data', [b'9', 25])
    def test_str_encode_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            str_enc = mdsEnc.Str()
            _ = str_enc.encode(data)

    @pytest.mark.parametrize(('data', 'encode_data'), [(99, b'c\x00\x00\x00\x00\x00\x00\x00'),
                                                       (987654321, b'\xb1h\xde:\x00\x00\x00\x00')])
    def test_int_encode_decode(self, data: int, encode_data: bytes):
        int_enc = mdsEnc.Int()
        assert int_enc.size == 8

        # Test encode
        enc_data = int_enc.encode(data)
        assert isinstance(enc_data, bytes)
        assert enc_data == encode_data

        # Test decode
        dec_data = int_enc.decode(encode_data)
        assert isinstance(dec_data, int)
        assert dec_data == data

    @pytest.mark.parametrize('data', [b'9', 25.9])
    def test_int_encode_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            int_enc = mdsEnc.Int()
            _ = int_enc.encode(data)

    @pytest.mark.parametrize('dtype_str', [
        'uint8',
        'uint16',
        'uint32',
        'uint64',
        'int8',
        'int16',
        'int32',
        'int64',
        'float16',
        'float32',
        'float64',
    ])
    @pytest.mark.parametrize('shape', [
        (1,),
        (1, 1),
        (2, 3, 4, 5),
        (1, 3, 1, 3, 1),
        (300,),
        (1, 256, 1),
        (65536, 7, 1),
    ])
    def test_ndarray_encode_decode(self, dtype_str: str, shape: tuple[int]):
        dtype = getattr(np, dtype_str)
        a = np.random.randint(0, 1000, shape).astype(dtype)

        encoding = 'ndarray'
        assert mdsEnc.is_mds_encoding(encoding)
        assert mdsEnc.get_mds_encoded_size(encoding) is None
        b = mdsEnc.mds_encode(encoding, a)
        c = mdsEnc.mds_decode(encoding, b)
        assert (a == c).all()
        b1_len = len(b)

        encoding = f'ndarray:{dtype.__name__}'
        assert mdsEnc.is_mds_encoding(encoding)
        assert mdsEnc.get_mds_encoded_size(encoding) is None
        b = mdsEnc.mds_encode(encoding, a)
        c = mdsEnc.mds_decode(encoding, b)
        assert (a == c).all()
        b2_len = len(b)

        shape_str = ','.join(map(str, shape))
        encoding = f'ndarray:{dtype.__name__}:{shape_str}'
        assert mdsEnc.is_mds_encoding(encoding)
        b_size = mdsEnc.get_mds_encoded_size(encoding)
        assert b_size is not None
        b = mdsEnc.mds_encode(encoding, a)
        c = mdsEnc.mds_decode(encoding, b)
        assert (a == c).all()
        assert len(b) == b_size
        b3_len = len(b)

        assert b3_len < b2_len < b1_len
        assert b3_len == np.prod(shape) * dtype().nbytes

    def test_error_no_elements_ndarray(self):
        encoding = 'ndarray'
        with pytest.raises(ValueError,
                           match='Attempting to encode a numpy array with 0 elements.*'):
            _ = mdsEnc.mds_encode(encoding, np.array([]))

    @pytest.mark.parametrize('array', [np.array(0.5), np.empty(()), np.array(1)])
    def test_error_scalar_ndarray(self, array: NDArray):
        encoding = 'ndarray'
        with pytest.raises(ValueError,
                           match='Attempting to encode a scalar with NDArray encoding.*'):
            _ = mdsEnc.mds_encode(encoding, array)

    @pytest.mark.parametrize('mode', ['I', 'L', 'RGB'])
    def test_pil_encode_decode(self, mode: str):
        pil_enc = mdsEnc.PIL()
        assert pil_enc.size is None

        # Creating the (32 x 32) NumPy Array with random values
        np_data = np.random.randint(255, size=(32, 32), dtype=np.uint32)
        # Default image mode of PIL Image is 'I'
        img = Image.fromarray(np_data).convert(mode)

        # Test encode
        enc_data = pil_enc.encode(img)
        assert isinstance(enc_data, bytes)

        # Test decode
        dec_data = pil_enc.decode(enc_data)
        dec_data = dec_data.convert('I')
        np_dec_data = np.asarray(dec_data, dtype=np.uint32)
        assert isinstance(dec_data, Image.Image)

        # Validate data content
        assert np.array_equal(np_data, np_dec_data)

    @pytest.mark.parametrize('data', [b'9', 25.9])
    def test_pil_encode_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            pil_enc = mdsEnc.PIL()
            _ = pil_enc.encode(data)

    @pytest.mark.parametrize('mode', ['L', 'RGB'])
    def test_jpeg_encode_decode(self, mode: str):
        jpeg_enc = mdsEnc.JPEG()
        assert jpeg_enc.size is None

        # Creating the (32 x 32) NumPy Array with random values
        np_data = np.random.randint(255, size=(32, 32), dtype=np.uint32)
        # Default image mode of PIL Image is 'I'
        img = Image.fromarray(np_data).convert(mode)

        # Test encode
        enc_data = jpeg_enc.encode(img)
        assert isinstance(enc_data, bytes)

        # Test decode
        dec_data = jpeg_enc.decode(enc_data)
        dec_data = dec_data.convert('I')
        assert isinstance(dec_data, Image.Image)

    @pytest.mark.parametrize('mode', ['L', 'RGB'])
    def test_jpegfile_encode_decode(self, mode: str):
        jpeg_enc = mdsEnc.JPEG()
        assert jpeg_enc.size is None

        # Creating the (32 x 32) NumPy Array with random values
        size = {'RGB': (224, 224, 3), 'L': (28, 28)}[mode]
        np_data = np.array(np.random.randint(255, size=size, dtype=np.uint8))
        # Default image mode of PIL Image is 'I'
        img = Image.fromarray(np_data).convert(mode)

        with tempfile.NamedTemporaryFile('wb') as f:
            img.save(f, format='jpeg')
            img = Image.open(f.name)

            # Test encode
            enc_data = jpeg_enc.encode(img)
            assert isinstance(enc_data, bytes)

            # Test decode
            dec_data = jpeg_enc.decode(enc_data)
            assert isinstance(dec_data, Image.Image)

            assert np.array_equal(np.array(img), np.array(dec_data))

    @pytest.mark.parametrize('data', [b'99', 12.5])
    def test_jpeg_encode_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            jpeg_enc = mdsEnc.JPEG()
            _ = jpeg_enc.encode(data)

    @pytest.mark.parametrize('mode', ['I', 'L', 'RGB'])
    def test_png_encode_decode(self, mode: str):
        png_enc = mdsEnc.PNG()
        assert png_enc.size is None

        # Creating the (32 x 32) NumPy Array with random values
        np_data = np.random.randint(255, size=(32, 32), dtype=np.uint32)
        # Default image mode of PIL Image is 'I'
        img = Image.fromarray(np_data).convert(mode)

        # Test encode
        enc_data = png_enc.encode(img)
        assert isinstance(enc_data, bytes)

        # Test decode
        dec_data = png_enc.decode(enc_data)
        dec_data = dec_data.convert('I')
        np_dec_data = np.asarray(dec_data, dtype=np.uint32)
        assert isinstance(dec_data, Image.Image)

        # Validate data content
        assert np.array_equal(np_data, np_dec_data)

    @pytest.mark.parametrize('data', [b'123', 77.7])
    def test_png_encode_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            png_enc = mdsEnc.PNG()
            _ = png_enc.encode(data)

    @pytest.mark.parametrize('data', [25, 'streaming', np.array(7)])
    def test_pickle_encode_decode(self, data: Any):
        pkl_enc = mdsEnc.Pickle()
        assert pkl_enc.size is None

        # Test encode
        enc_data = pkl_enc.encode(data)
        assert isinstance(enc_data, bytes)

        # Test decode
        dec_data = pkl_enc.decode(enc_data)
        assert isinstance(dec_data, type(data))

        # Validate data content
        assert dec_data == data

    @pytest.mark.parametrize('data', [25, 'streaming', {'alpha': 1, 'beta': 2}])
    def test_json_encode_decode(self, data: Any):
        json_enc = mdsEnc.JSON()
        assert json_enc.size is None

        # Test encode
        enc_data = json_enc.encode(data)
        assert isinstance(enc_data, bytes)

        # Test decode
        dec_data = json_enc.decode(enc_data)
        assert isinstance(dec_data, type(data))

        # Validate data content
        assert dec_data == data

    @pytest.mark.parametrize('data', [np.array([1]), np.array(['foo']), np.array([{'foo': 1}])])
    def test_json_encode_decode_ndarray(self, data: Any):
        json_enc = mdsEnc.JSON()
        assert json_enc.size is None

        # Test encode
        enc_data = json_enc.encode(data)
        assert isinstance(enc_data, bytes)

        # Test decode
        dec_data = json_enc.decode(enc_data)
        assert isinstance(dec_data, list)

        # Validate data content
        assert dec_data == data.tolist()

    def test_json_invalid_data(self):
        wrong_json_with_single_quotes = "{'name': 'streaming'}"
        with pytest.raises(json.JSONDecodeError):
            json_enc = mdsEnc.JSON()
            json_enc._is_valid(wrong_json_with_single_quotes, wrong_json_with_single_quotes)

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'*')])
    def test_mds_uint8(self, decoded: int, encoded: bytes):
        coder = mdsEnc.UInt8()
        assert coder.size == 1

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.integer)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'*\0')])
    def test_mds_uint16(self, decoded: int, encoded: bytes):
        coder = mdsEnc.UInt16()
        assert coder.size == 2

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.integer)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'*\0\0\0')])
    def test_mds_uint32(self, decoded: int, encoded: bytes):
        coder = mdsEnc.UInt32()
        assert coder.size == 4

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.integer)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'*\0\0\0\0\0\0\0')])
    def test_mds_uint64(self, decoded: int, encoded: bytes):
        coder = mdsEnc.UInt64()
        assert coder.size == 8

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.integer)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'*')])
    def test_mds_int8(self, decoded: int, encoded: bytes):
        coder = mdsEnc.Int8()
        assert coder.size == 1

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.integer)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'*\0')])
    def test_mds_int16(self, decoded: int, encoded: bytes):
        coder = mdsEnc.Int16()
        assert coder.size == 2

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.integer)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'*\0\0\0')])
    def test_mds_int32(self, decoded: int, encoded: bytes):
        coder = mdsEnc.Int32()
        assert coder.size == 4

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.integer)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'*\0\0\0\0\0\0\0')])
    def test_mds_int64(self, decoded: int, encoded: bytes):
        coder = mdsEnc.Int64()
        assert coder.size == 8

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.integer)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42.0, b'@Q')])
    def test_mds_float16(self, decoded: float, encoded: bytes):
        coder = mdsEnc.Float16()
        assert coder.size == 2

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.floating)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42.0, b'\0\0(B')])
    def test_mds_float32(self, decoded: float, encoded: bytes):
        coder = mdsEnc.Float32()
        assert coder.size == 4

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.floating)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42.0, b'\0\0\0\0\0\0E@')])
    def test_mds_float64(self, decoded: float, encoded: bytes):
        coder = mdsEnc.Float64()
        assert coder.size == 8

        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, np.floating)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42, b'42'), (-42, b'-42')])
    def test_mds_StrInt(self, decoded: int, encoded: bytes):
        coder = mdsEnc.StrInt()
        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, int)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(42.0, b'42.0'), (-42.0, b'-42.0')])
    def test_mds_StrFloat(self, decoded: float, encoded: bytes):
        coder = mdsEnc.StrFloat()
        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, float)
        assert dec == decoded

    @pytest.mark.parametrize(('decoded', 'encoded'), [(Decimal('4E15'), b'4E+15'),
                                                      (Decimal('-4E15'), b'-4E+15')])
    def test_mds_StrDecimal(self, decoded: Decimal, encoded: bytes):
        coder = mdsEnc.StrDecimal()
        enc = coder.encode(decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded

        dec = coder.decode(encoded)
        assert isinstance(dec, Decimal)
        assert dec == decoded

    def test_get_mds_encodings(self):
        uints = {'uint8', 'uint16', 'uint32', 'uint64'}
        ints = {'int8', 'int16', 'int32', 'int64', 'str_int'}
        floats = {'float16', 'float32', 'float64', 'str_float', 'str_decimal'}
        scalars = uints | ints | floats
        expected_encodings = {
            'int', 'bytes', 'json', 'ndarray', 'png', 'jpeg', 'str', 'pil', 'pkl'
        } | scalars
        enc = mdsEnc.get_mds_encodings()
        assert len(enc) == len(expected_encodings)
        assert enc == expected_encodings

    @pytest.mark.parametrize(('enc_name', 'expected_output'), [('jpeg', True), ('', False),
                                                               ('pngg', False)])
    def test_is_mds_encoding(self, enc_name: str, expected_output: bool):
        is_supported = mdsEnc.is_mds_encoding(enc_name)
        assert is_supported is expected_output

    @pytest.mark.parametrize(('encoding', 'decoded', 'encoded'),
                             [('uint8', 42, b'*'), ('uint16', 42, b'*\0'),
                              ('uint32', 42, b'*\0\0\0'), ('uint64', 42, b'*\0\0\0\0\0\0\0'),
                              ('int8', 42, b'*'), ('int16', 42, b'*\0'), ('int32', 42, b'*\0\0\0'),
                              ('int64', 42, b'*\0\0\0\0\0\0\0'), ('float16', 42.0, b'@Q'),
                              ('float32', 42.0, b'\0\0(B'), ('float64', 42.0, b'\0\0\0\0\0\0E@')])
    def test_mds_scalar(self, encoding: str, decoded: Union[int, float], encoded: bytes):
        enc = mdsEnc.mds_encode(encoding, decoded)
        assert isinstance(enc, bytes)
        assert enc == encoded
        dec = mdsEnc.mds_decode(encoding, enc)
        assert dec == decoded
        dec = mdsEnc.mds_decode(encoding, encoded)
        assert dec == decoded

    @pytest.mark.parametrize(('enc_name', 'data'), [('bytes', b'9'), ('int', 27),
                                                    ('str', 'mosaicml')])
    def test_mds_encode(self, enc_name: str, data: Any):
        output = mdsEnc.mds_encode(enc_name, data)
        assert isinstance(output, bytes)

    @pytest.mark.parametrize(('enc_name', 'data'), [('bytes', 9), ('int', '27'), ('str', 12.5)])
    def test_mds_encode_invalid_data(self, enc_name: str, data: Any):
        with pytest.raises(AttributeError):
            _ = mdsEnc.mds_encode(enc_name, data)

    @pytest.mark.parametrize(('enc_name', 'data', 'expected_data_type'),
                             [('bytes', b'c\x00\x00\x00\x00\x00\x00\x00', bytes),
                              ('str', b'mosaicml', str)])
    def test_mds_decode(self, enc_name: str, data: Any, expected_data_type: Any):
        output = mdsEnc.mds_decode(enc_name, data)
        assert isinstance(output, expected_data_type)

    @pytest.mark.parametrize(('enc_name', 'expected_size'), [('bytes', None), ('int', 8)])
    def test_get_mds_encoded_size(self, enc_name: str, expected_size: Any):
        output = mdsEnc.get_mds_encoded_size(enc_name)
        assert output is expected_size


class TestXSVEncodings:

    @pytest.mark.parametrize(('data', 'encode_data'), [('99', '99'),
                                                       ('streaming dataset', 'streaming dataset')])
    def test_str_encode_decode(self, data: str, encode_data: str):
        str_enc = xsvEnc.Str()

        # Test encode
        enc_data = str_enc.encode(data)
        assert isinstance(enc_data, str)
        assert enc_data == encode_data

        # Test decode
        dec_data = str_enc.decode(encode_data)
        assert isinstance(dec_data, str)
        assert dec_data == data

    @pytest.mark.parametrize('data', [99, b'streaming dataset', 123.45])
    def test_str_encode_invalid_data(self, data: Any):
        with pytest.raises(Exception):
            str_enc = xsvEnc.Str()
            _ = str_enc.encode(data)

    @pytest.mark.parametrize(('data', 'encode_data'), [(99, '99'), (987675432, '987675432')])
    def test_int_encode_decode(self, data: int, encode_data: str):
        int_enc = xsvEnc.Int()

        # Test encode
        enc_data = int_enc.encode(data)
        assert isinstance(enc_data, str)
        assert enc_data == encode_data

        # Test decode
        dec_data = int_enc.decode(encode_data)
        assert isinstance(dec_data, int)
        assert dec_data == data

    @pytest.mark.parametrize('data', ['99', b'streaming dataset', 123.45])
    def test_int_encode_invalid_data(self, data: Any):
        with pytest.raises(Exception):
            int_enc = xsvEnc.Int()
            _ = int_enc.encode(data)

    @pytest.mark.parametrize(('data', 'encode_data'), [(1.24, '1.24'), (9.0, '9.0')])
    def test_float_encode_decode(self, data: int, encode_data: str):
        float_enc = xsvEnc.Float()

        # Test encode
        enc_data = float_enc.encode(data)
        assert isinstance(enc_data, str)
        assert enc_data == encode_data

        # Test decode
        dec_data = float_enc.decode(encode_data)
        assert isinstance(dec_data, float)
        assert dec_data == data

    @pytest.mark.parametrize('data', ['99', b'streaming dataset', 12])
    def test_float_encode_invalid_data(self, data: Any):
        with pytest.raises(Exception):
            float_enc = xsvEnc.Float()
            _ = float_enc.encode(data)

    @pytest.mark.parametrize(('enc_name', 'expected_output'), [
        ('str', True),
        ('int', True),
        ('float', True),
        ('', False),
    ])
    def test_is_xsv_encoding(self, enc_name: str, expected_output: bool):
        is_supported = xsvEnc.is_xsv_encoding(enc_name)
        assert is_supported is expected_output

    @pytest.mark.parametrize(('enc_name', 'data', 'expected_data'),
                             [('str', 'mosaicml', 'mosaicml'), ('int', 27, '27'),
                              ('float', 1.25, '1.25')])
    def test_xsv_encode(self, enc_name: str, data: Any, expected_data: str):
        output = xsvEnc.xsv_encode(enc_name, data)
        assert isinstance(output, str)
        assert output == expected_data

    @pytest.mark.parametrize(('enc_name', 'data', 'expected_data'),
                             [('str', 'mosaicml', 'mosaicml'), ('int', '27', 27),
                              ('float', '1.25', 1.25)])
    def test_xsv_decode(self, enc_name: str, data: str, expected_data: Any):
        output = xsvEnc.xsv_decode(enc_name, data)
        assert isinstance(output, type(expected_data))
        assert output == expected_data


class TestJSONEncodings:

    @pytest.mark.parametrize('data', ['99', 'mosaicml'])
    def test_str_is_encoded(self, data: str):
        json_enc = jsonEnc.Str()

        # Test encode
        enc_data = json_enc.is_encoded(data)
        assert isinstance(enc_data, bool)

    @pytest.mark.parametrize('data', [99, b'mosaicml'])
    def test_str_is_encoded_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            json_enc = jsonEnc.Str()
            _ = json_enc.is_encoded(data)

    @pytest.mark.parametrize('data', [99, 987675432])
    def test_int_is_encoded(self, data: int):
        int_enc = jsonEnc.Int()

        # Test encode
        enc_data = int_enc.is_encoded(data)
        assert isinstance(enc_data, bool)

    @pytest.mark.parametrize('data', ['99', b'mosaicml', 1.25])
    def test_int_is_encoded_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            int_enc = jsonEnc.Int()
            _ = int_enc.is_encoded(data)

    @pytest.mark.parametrize('data', [1.25])
    def test_float_is_encoded(self, data: int):
        float_enc = jsonEnc.Float()

        # Test encode
        enc_data = float_enc.is_encoded(data)
        assert isinstance(enc_data, bool)

    @pytest.mark.parametrize('data', ['99', b'mosaicml', 25])
    def test_float_is_encoded_invalid_data(self, data: Any):
        with pytest.raises(AttributeError):
            float_enc = jsonEnc.Float()
            _ = float_enc.is_encoded(data)

    @pytest.mark.parametrize(('enc_name', 'expected_output'), [
        ('str', True),
        ('int', True),
        ('float', True),
        ('', False),
    ])
    def test_is_json_encoding(self, enc_name: str, expected_output: bool):
        is_supported = jsonEnc.is_json_encoding(enc_name)
        assert is_supported is expected_output

    @pytest.mark.parametrize(('enc_name', 'data', 'expected_output'), [('str', 'hello', True),
                                                                       ('int', 10, True),
                                                                       ('float', 9.9, True)])
    def test_is_json_encoded(self, enc_name: str, data: Any, expected_output: bool):
        is_supported = jsonEnc.is_json_encoded(enc_name, data)
        assert is_supported is expected_output



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

from typing import Any

import pytest

from streaming import MDSWriter, StreamingDataset


def one(remote: str, local: str):
    """
    Verify __getitem__ accesses.
    """
    dataset = StreamingDataset(local=remote, batch_size=1)
    for i in range(dataset.num_samples):
        sample = dataset[i]
        assert sample['value'] == i


def two(remote: str, local: str):
    """
    Verify __iter__ -> __getitem__ accesses.
    """
    dataset = StreamingDataset(local=remote, num_canonical_nodes=1, batch_size=1)
    for i, sample in zip(range(dataset.num_samples), dataset):
        assert sample['value'] == i


def three(remote: str, local: str):
    """
    Verify __getitem__ downloads/accesses.
    """
    dataset = StreamingDataset(local=local, remote=remote, batch_size=1)
    for i in range(dataset.num_samples):
        sample = dataset[i]
        assert sample['value'] == i


def four(remote: str, local: str):
    """
    Verify __iter__ -> __getitem__ downloads/accesses.
    """
    dataset = StreamingDataset(local=local, remote=remote, num_canonical_nodes=1, batch_size=1)
    for i, sample in zip(range(dataset.num_samples), dataset):
        assert sample['value'] == i
    del dataset


def five(remote: str, local: str):
    """
    Re-verify __getitem__ downloads/accesses.
    """
    dataset = StreamingDataset(local=local, remote=remote, batch_size=1)
    for i in range(dataset.num_samples):
        sample = dataset[i]
        assert sample['value'] == i


@pytest.mark.usefixtures('local_remote_dir')
@pytest.mark.parametrize('func', [one, two, three, four, five])
def test_laziness(local_remote_dir: tuple[str, str], func: Any):
    num_samples = 10_000
    local, remote = local_remote_dir
    columns = {'value': 'int'}
    compression = None
    hashes = None
    size_limit = 1_000

    with MDSWriter(out=remote,
                   columns=columns,
                   compression=compression,
                   hashes=hashes,
                   size_limit=size_limit) as out:
        for i in range(num_samples):
            sample = {'value': i}
            out.write(sample)

    func(remote, local)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Class for tracking node information during simulation."""

from typing import Optional

import numpy as np
from core.last_used_ordered_set import LastUsedOrderedSet
from core.utils import remove_padded_samples
from numpy.typing import NDArray
from sortedcollections import OrderedSet

from streaming.base.spanner import Spanner


class NodeTracker():
    """Tracker for node information during simulation.

    Args:
        workers (int): The number of workers.
        devices (int): The number of devices.
        predownload (int): The number of samples to predownload.
        device_batch_size (int): The device batch size.
        total_shards (int): Total number of shards in the dataset.
        cache_limit (Optional[int]): The cache limit for the node. Defaults to ``None``.
    """

    def __init__(self,
                 workers: int,
                 devices: int,
                 predownload: int,
                 device_batch_size: int,
                 total_shards: int,
                 cache_limit: Optional[int] = None):

        self.shards = LastUsedOrderedSet()
        self.all_shards = set()
        self.cache_usage = 0
        self.partial_shard_bytes = 0
        self.worker_download_index = 0
        self.devices = devices
        self.workers = workers
        self.total_workers = workers * devices
        self.device_batch_size = device_batch_size
        self.predownload = predownload
        self.cache_limit = cache_limit
        self.worker_downloads = []
        self.shard_access_starts = np.full(total_shards, -1)
        self.shard_access_ends = np.full(total_shards, -1)

        # Use the set_epoch_samples method every epoch to set the node's samples.
        self.samples = None

    def initialize_worker_downloads(self, sample_to_shard: Spanner):
        """Initialize worker downloads, making shards in the predownload sample range available.

        Args:
            sample_to_shard (Spanner): The mapping from samples to shards.
        """
        # For downloads, we round-robin over devices first, then workers.
        if self.samples is not None:
            for worker in range(self.workers):
                for device in range(self.devices):
                    download_samples = remove_padded_samples(
                        self.samples[device, worker, :self.predownload])
                    # Get the shards these samples correspond to, maintaining access order.
                    download_shards = OrderedSet(
                        [sample_to_shard[sample] for sample in download_samples])
                    self.worker_downloads.append(download_shards)
        else:
            raise AttributeError('Must set node samples before accessing them.')

    def set_shards_used(self, shards: set, step_num: int):
        """Mark a set of shards as recently used.

        Args:
            shards (set): The shards to set as used.
            step_num (int): The current step number.
        """
        for shard in shards:
            self.shards.setuse(shard)
            # For any shard access, we are accessing the shard so we need the shard until
            # at least the next step begins. Adding 0.5 ensures that we evict shards
            # after they are used for the last time, but before they are replaced by
            # new downloads in the next step.
            self.shard_access_ends[shard] = step_num + 0.5

    def add_shard(self, shard: int, used: bool = True):
        """Add a shard to the node.

        Args:
            shard (int): The shard to add.
            used (bool): Whether the shard is used when added.
        """
        self.shards.setitem(shard, used)
        self.all_shards.add(shard)

    def get_all_shards(self):
        """Get all the shards in the node."""
        return self.all_shards

    def evict_shard(self) -> int:
        """Evict a shard.

        Returns:
            int: The evicted shard.
        """
        evicted_shard = self.shards.popLRU()
        return evicted_shard

    def evict_until_satisfied(self, incoming_shard_size: int, raw_shard_sizes: NDArray[np.int64]):
        """Evict shards until the node has enough space to download the incoming shard.

        Args:
            incoming_shard_size (int): The size of the incoming shard.
            raw_shard_sizes (NDArray[np.int64]): The raw shard sizes.
        """
        # We evict shards until the incoming shard fits into the node's cache.
        if self.cache_limit is not None:
            while self.cache_usage + incoming_shard_size > self.cache_limit:
                evicted_shard = self.evict_shard()
                self.cache_usage -= int(raw_shard_sizes[evicted_shard])

    def increment_worker_download_index(self):
        """Increment the worker download index."""
        self.worker_download_index = (self.worker_download_index + 1) % \
            (self.workers * self.devices)

    def get_worker_download(self,
                            worker: Optional[int] = None,
                            device: Optional[int] = None,
                            index: Optional[int] = None) -> OrderedSet:
        """Get the shard downloads for a worker on a specific device.

        Args:
            worker (Optional[int]): The worker index.
            device (Optional[int]): The device index the worker is on.
            index (Optional[int]): The index of the worker download for direct access.

        Returns:
            OrderedSet: The shard downloads, in order, for this worker.
        """
        if index is not None:
            # Directly access worker_downloads through an index.
            return self.worker_downloads[index]
        elif worker is not None and device is not None:
            # Access worker_downloads through worker and device indices.
            return self.worker_downloads[worker * self.devices + device]
        else:
            raise ValueError('Must specify either index, or worker and device.')

    def get_current_batch_shards(self, worker: int, worker_sample_index: int,
                                 sample_to_shard: Spanner) -> tuple[set, set]:
        """Get this node's shards for the current batch.

        Args:
            worker (int): The worker.
            worker_sample_index (int): The worker sample index.
            sample_to_shard (Spanner): The mapping from samples to shards.

        Returns:
            Tuple[set, set]: shard ids needed by node, shard ids present in node.
        """
        if self.samples is not None:
            batch_samples = remove_padded_samples(
                self.samples[:, worker, worker_sample_index:worker_sample_index +
                             self.device_batch_size].flatten())
            batch_shards = {sample_to_shard[sample] for sample in batch_samples}
            shards_needed = batch_shards.difference(self.shards.keys())
            shards_present = batch_shards.difference(shards_needed)
            return shards_needed, shards_present
        else:
            raise AttributeError('Must set node samples before accessing them.')

    def get_next_worker_with_downloads(self) -> Optional[OrderedSet]:
        """Get the next worker with samples to download.

        Returns:
            Optional[OrderedSet]: The next worker's shard downloads, or None if no workers have
                samples to download.
        """
        empty_download_counter = 0
        worker_download = self.get_worker_download(index=self.worker_download_index)
        while len(worker_download) == 0:
            empty_download_counter += 1
            self.increment_worker_download_index()
            worker_download = self.get_worker_download(index=self.worker_download_index)

            # No workers in the node have samples to download.
            if empty_download_counter >= self.total_workers:
                return None

        return worker_download

    def update_worker_predownloads(self, worker: int, worker_sample_index: int,
                                   sample_to_shard: Spanner):
        """Update the worker predownload samples for a worker and device.

        Args:
            worker (int): The current batch worker index.
            worker_sample_index (int): The worker sample index.
            sample_to_shard (Spanner): The mapping from samples to shards.
        """
        if self.samples is not None:
            for device in range(self.devices):
                # Retrieve new samples that are now within predownload range of the worker.
                new_download_samples = remove_padded_samples(
                    self.samples[device, worker,
                                 worker_sample_index + self.predownload:worker_sample_index +
                                 self.device_batch_size + self.predownload])

                # Want to maintain the shard access order.
                new_download_shards = OrderedSet(
                    [sample_to_shard[sample] for sample in new_download_samples])

                worker_downloads = self.get_worker_download(worker=worker, device=device)

                # Add in new shards to the worker's shard downloads only if node does not yet have it.
                for shard in new_download_shards:
                    if shard not in self.shards:
                        worker_downloads.add(shard)
        else:
            raise AttributeError('Must set node samples before accessing them.')



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import operator
import os
from shutil import rmtree
from typing import Any

import pytest
from torch.utils.data import DataLoader

from streaming import MDSWriter, StreamingDataset
from tests.common.utils import convert_to_mds


def validate(remote: str, local: str, dataset: StreamingDataset, keep_zip: bool,
             is_shard_evicted: bool):
    """Validate the number of files in a local directory in comparison to remote directory."""
    if is_shard_evicted:
        ops = operator.lt
    else:
        ops = operator.eq
    if keep_zip:
        if dataset.shards[0].compression:
            # Local has raw + zip, remote has zip.
            assert ops(
                set(filter(lambda f: f == 'index.json' or f.endswith('.zstd'), os.listdir(local))),
                set(os.listdir(remote)))
        else:
            # Local has raw + zip, remote has raw.
            assert ops(set(filter(lambda f: not f.endswith('.zstd'), os.listdir(local))),  \
                set(os.listdir(remote)))
    else:
        if dataset.shards[0].compression:
            # Local has raw, remote has zip.
            assert ops(set(os.listdir(local)),
                       {f.replace('.zstd', '') for f in os.listdir(remote)})
        else:
            # Local has raw, remote has raw.
            assert ops(set(os.listdir(local)), set(os.listdir(remote)))


def shard_eviction_disabled(remote: str, local: str, keep_zip: bool):
    """
    With shard eviction disabled.
    """
    dataset = StreamingDataset(remote=remote, local=local, keep_zip=keep_zip, batch_size=1)
    for _ in range(2):
        for sample in dataset:  # pyright: ignore
            pass

    validate(remote, local, dataset, keep_zip, False)
    rmtree(local, ignore_errors=False)


def shard_eviction_too_high(remote: str, local: str, keep_zip: bool):
    """
    With no shard evictions because cache_limit is bigger than the dataset.
    """
    dataset = StreamingDataset(remote=remote,
                               local=local,
                               keep_zip=keep_zip,
                               cache_limit=1_000_000,
                               batch_size=1)
    dataloader = DataLoader(dataset=dataset, num_workers=8)
    for _ in range(2):
        for _ in dataloader:
            pass
    validate(remote, local, dataset, keep_zip, False)
    rmtree(local, ignore_errors=False)


def shard_eviction(remote: str, local: str, keep_zip: bool):
    """
    With shard eviction because cache_limit is smaller than the whole dataset.
    """
    cache_limit = '120kb' if keep_zip else '100kb'
    dataset = StreamingDataset(remote=remote,
                               local=local,
                               keep_zip=keep_zip,
                               cache_limit=cache_limit,
                               batch_size=1)
    dataloader = DataLoader(dataset=dataset, num_workers=8)
    for _ in range(2):
        for _ in dataloader:
            pass
    validate(remote, local, dataset, keep_zip, True)
    rmtree(local, ignore_errors=False)


def manual_shard_eviction(remote: str, local: str, keep_zip: bool):
    """
    Manually downloading and evicting shards.
    """
    dataset = StreamingDataset(remote=remote, local=local, keep_zip=keep_zip, batch_size=1)

    for shard_id in range(dataset.num_shards):
        dataset.prepare_shard(shard_id)

    full = set(os.listdir(local))

    for shard_id in range(dataset.num_shards):
        dataset.evict_shard(shard_id)

    assert os.listdir(local) == ['index.json']

    for sample_id in range(dataset.num_samples):
        dataset[sample_id]

    assert set(os.listdir(local)) == full
    rmtree(local, ignore_errors=False)


def cache_limit_too_low(remote: str, local: str, keep_zip: bool):
    """
    With impossible shard eviction settings because cache_limit is set too low.
    """
    with pytest.raises(ValueError):
        dataset = StreamingDataset(remote=remote, local=local, cache_limit='1kb', batch_size=1)
        for _ in dataset:
            pass
    rmtree(local, ignore_errors=False)


funcs = [
    shard_eviction_disabled, shard_eviction_too_high, shard_eviction, manual_shard_eviction,
    cache_limit_too_low
]


@pytest.mark.usefixtures('local_remote_dir')
@pytest.mark.parametrize('func', list(funcs))
def test_eviction_nozip(local_remote_dir: tuple[str, str], func: Any):
    num_samples = 5_000
    local, remote = local_remote_dir
    columns = {'data': 'bytes'}
    compression = None
    hashes = None
    size_limit = 500

    with MDSWriter(out=remote,
                   columns=columns,
                   compression=compression,
                   hashes=hashes,
                   size_limit=size_limit) as out:
        for _ in range(num_samples):
            sample = {'data': b'\0'}
            out.write(sample)

    func(remote, local, False)


@pytest.mark.usefixtures('local_remote_dir')
@pytest.mark.parametrize('func', list(funcs))
def test_eviction_zip_nokeep(local_remote_dir: tuple[str, str], func: Any):
    num_samples = 5_000
    local, remote = local_remote_dir
    columns = {'data': 'bytes'}
    compression = 'zstd'
    hashes = None
    size_limit = 500

    with MDSWriter(out=remote,
                   columns=columns,
                   compression=compression,
                   hashes=hashes,
                   size_limit=size_limit) as out:
        for _ in range(num_samples):
            sample = {'data': b'\0'}
            out.write(sample)

    func(remote, local, False)


@pytest.mark.usefixtures('local_remote_dir')
@pytest.mark.parametrize('func', list(funcs))
def test_eviction_zip_keep(local_remote_dir: tuple[str, str], func: Any):
    num_samples = 5_000
    local, remote = local_remote_dir
    columns = {'data': 'bytes'}
    compression = 'zstd'
    hashes = None
    size_limit = 500

    with MDSWriter(out=remote,
                   columns=columns,
                   compression=compression,
                   hashes=hashes,
                   size_limit=size_limit) as out:
        for _ in range(num_samples):
            sample = {'data': b'\0'}
            out.write(sample)

    func(remote, local, True)


@pytest.mark.parametrize('cache_limit', ['2048'])
@pytest.mark.usefixtures('local_remote_dir')
def test_cache_limit_lower_than_index_json(local_remote_dir: Any, cache_limit: str):
    remote_dir, local_dir = local_remote_dir
    convert_to_mds(out_root=remote_dir,
                   dataset_name='sequencedataset',
                   num_samples=1000,
                   compression=None,
                   size_limit=2048)

    with pytest.raises(ValueError, match='Minimum cache usage.*is larger than the cache limit*'):
        _ = StreamingDataset(local=local_dir,
                             remote=remote_dir,
                             shuffle=False,
                             batch_size=4,
                             cache_limit=cache_limit)


@pytest.mark.parametrize('cache_limit', ['5kb'])
@pytest.mark.parametrize('compression', [None, 'zstd:7'])
@pytest.mark.usefixtures('local_remote_dir')
def test_cache_limit_lower_than_few_shards(local_remote_dir: Any, cache_limit: str,
                                           compression: str):
    remote_dir, local_dir = local_remote_dir
    convert_to_mds(out_root=remote_dir,
                   dataset_name='sequencedataset',
                   num_samples=1000,
                   compression=compression,
                   size_limit=2048)

    with pytest.raises(ValueError,
                       match='Cache limit.*is too low. Increase the `cache_limit` to*'):
        _ = StreamingDataset(local=local_dir,
                             remote=remote_dir,
                             shuffle=False,
                             batch_size=4,
                             cache_limit=cache_limit)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

""":class:`MDSReader` reads samples in `.mds` files written by :class:`StreamingDatasetWriter`."""

import os
from copy import deepcopy
from typing import Any, Optional, Union

import numpy as np
from typing_extensions import Self

from streaming.base.format.base.reader import FileInfo, JointReader
from streaming.base.format.mds.encodings import is_mds_encoding_safe, mds_decode

__all__ = ['MDSReader']


class MDSReader(JointReader):
    """Provides random access to the samples of an MDS shard.

    Args:
        dirname (str): Local dataset directory.
        split (str, optional): Which dataset split to use, if any.
        column_encodings (List[str]): Column encodings.
        column_names (List[str]): Column names.
        column_sizes (List[Optional[int]]): Column fixed sizes, if any.
        compression (str, optional): Optional compression or compression:level.
        hashes (List[str]): Optional list of hash algorithms to apply to shard files.
        raw_data (FileInfo): Uncompressed data file info.
        samples (int): Number of samples in this shard.
        size_limit (Union[int, str], optional): Optional shard size limit, after
            which point to start a new shard. If None, puts everything in one shard.
            Can specify bytes in human-readable format as well, for example
            ``"100kb"`` for 100 kilobyte (100*1024) and so on.
        zip_data (FileInfo, optional): Compressed data file info.
    """

    def __init__(
        self,
        dirname: str,
        split: Optional[str],
        column_encodings: list[str],
        column_names: list[str],
        column_sizes: list[Optional[int]],
        compression: Optional[str],
        hashes: list[str],
        raw_data: FileInfo,
        samples: int,
        size_limit: Optional[Union[int, str]],
        zip_data: Optional[FileInfo],
    ) -> None:
        super().__init__(dirname, split, compression, hashes, raw_data, samples, size_limit,
                         zip_data)
        self.column_encodings = column_encodings
        self.column_names = column_names
        self.column_sizes = column_sizes

    @classmethod
    def from_json(cls, dirname: str, split: Optional[str], obj: dict[str, Any]) -> Self:
        """Initialize from JSON object.

        Args:
            dirname (str): Local directory containing shards.
            split (str, optional): Which dataset split to use, if any.
            obj (Dict[str, Any]): JSON object to load.

        Returns:
            Self: Loaded MDSReader.
        """
        args = deepcopy(obj)
        if args['version'] != 2:
            raise ValueError(f'Unsupported streaming data version: {args["version"]}. ' +
                             f'Expected version 2.')
        del args['version']
        if args['format'] != 'mds':
            raise ValueError(f'Unsupported data format: {args["format"]}. ' +
                             f'Expected to be `mds`.')
        del args['format']
        args['dirname'] = dirname
        args['split'] = split
        for key in ['raw_data', 'zip_data']:
            arg = args[key]
            args[key] = FileInfo(**arg) if arg else None
        return cls(**args)

    def validate(self, allow_unsafe_types: bool) -> None:
        """Check whether this shard is acceptable to be part of some Stream.

        Args:
            allow_unsafe_types (bool): If a shard contains Pickle, which allows arbitrary code
                execution during deserialization, whether to keep going if ``True`` or raise an
                error if ``False``.
        """
        if not allow_unsafe_types:
            for column_id, encoding in enumerate(self.column_encodings):
                if not is_mds_encoding_safe(encoding):
                    name = self.column_names[column_id]
                    raise ValueError(f'Column {name} contains an unsafe type: {encoding}. To ' +
                                     f'proceed anyway, set ``allow_unsafe_types=True``.')

    def decode_sample(self, data: bytes) -> dict[str, Any]:
        """Decode a sample dict from bytes.

        Args:
            data (bytes): The sample encoded as bytes.

        Returns:
            Dict[str, Any]: Sample dict.
        """
        sizes = []
        idx = 0
        for key, size in zip(self.column_names, self.column_sizes):
            if size:
                sizes.append(size)
            else:
                size, = np.frombuffer(data[idx:idx + 4], np.uint32)
                sizes.append(size)
                idx += 4
        sample = {}
        for key, encoding, size in zip(self.column_names, self.column_encodings, sizes):
            value = data[idx:idx + size]
            sample[key] = mds_decode(encoding, value)
            idx += size
        return sample

    def get_sample_data(self, idx: int) -> bytes:
        """Get the raw sample data at the index.

        Args:
            idx (int): Sample index.

        Returns:
            bytes: Sample data.
        """
        filename = os.path.join(self.dirname, self.split, self.raw_data.basename)
        offset = (1 + idx) * 4
        with open(filename, 'rb', 0) as fp:
            fp.seek(offset)
            pair = fp.read(8)
            begin, end = np.frombuffer(pair, np.uint32)
            fp.seek(begin)
            data = fp.read(end - begin)
        if not data:
            raise IndexError(
                f'Relative sample index {idx} is not present in the {self.raw_data.basename} file.'
            )
        return data



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Utility function for converting spark dataframe to MDS dataset."""

from streaming.base.converters.dataframe_to_mds import (SPARK_TO_MDS, dataframe_to_mds,
                                                        dataframeToMDS, infer_dataframe_schema,
                                                        is_json_compatible)

__all__ = [
    'dataframeToMDS', 'dataframe_to_mds', 'SPARK_TO_MDS', 'infer_dataframe_schema',
    'is_json_compatible'
]



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import os
import shutil
import tempfile
from typing import Any
from unittest import mock
from unittest.mock import ANY, MagicMock, Mock, patch

import boto3
import pytest

from streaming.base.storage.upload import (AlipanUploader, AzureDataLakeUploader, AzureUploader,
                                           CloudUploader, DatabricksUnityCatalogUploader,
                                           DBFSUploader, GCSAuthentication, GCSUploader,
                                           HFUploader, LocalUploader, S3Uploader)
from tests.conftest import MY_BUCKET, R2_URL

MY_PREFIX = 'train'


@pytest.fixture(scope='function')
def remote_local_dir() -> Any:
    """Creates a temporary directory and then deletes it when the calling function is done."""

    def _method(cloud_prefix: str = '') -> tuple[str, str]:
        try:
            mock_local_dir = tempfile.TemporaryDirectory()
            mock_local = mock_local_dir.name
            mock_remote = os.path.join(cloud_prefix, MY_BUCKET, MY_PREFIX)
            return mock_remote, mock_local
        finally:
            mock_local_dir.cleanup()  # pyright: ignore

    return _method


class TestCloudUploader:

    @patch('streaming.base.storage.upload.S3Uploader.check_bucket_exists')
    @patch('streaming.base.storage.upload.GCSUploader.check_bucket_exists')
    @pytest.mark.parametrize(
        'mapping',
        [
            ['s3://bucket/dir/file', S3Uploader],
            [None, 's3://bucket/dir/file', S3Uploader],
            ['gs://bucket/dir/file', GCSUploader],
            [None, 'gs://bucket/dir/file', GCSUploader],
            ['/tmp/dir/filepath', LocalUploader],
            ['./relative/dir/filepath', LocalUploader],
        ],
    )
    @pytest.mark.usefixtures('gcs_hmac_credentials')
    def test_instantiation_type(
        self,
        s3_mocked_requests: Mock,
        gcs_mocked_requests: Mock,
        local_remote_dir: tuple[str, str],
        mapping: list[Any],
    ):
        s3_mocked_requests.side_effect = None
        gcs_mocked_requests.side_effect = None
        local, _ = local_remote_dir
        if len(mapping) == 2:
            cw = CloudUploader.get(out=mapping[0])
        else:
            mapping[0] = local
            out_root = (mapping[0], mapping[1])
            cw = CloudUploader.get(out_root)
        assert isinstance(cw, mapping[-1])

    @pytest.mark.parametrize('out', [(), ('s3://bucket/dir',), ('./dir1', './dir2', './dir3')])
    def test_invalid_out_parameter_length(self, out: Any):
        with pytest.raises(ValueError, match=f'Invalid `out` argument.*'):
            _ = CloudUploader.get(out=out)

    @pytest.mark.parametrize('out', [('./dir1', 'gcs://bucket/dir/'), ('./dir1', None)])
    def test_invalid_out_parameter_type(self, out: Any):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = CloudUploader.get(out=out)

    def test_local_directory_is_empty(self, local_remote_dir: tuple[str, str]):
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = CloudUploader.get(out=local)

    def test_local_directory_is_created(self, local_remote_dir: tuple[str, str]):
        local, _ = local_remote_dir
        _ = CloudUploader(out=local)
        assert os.path.exists(local)

    def test_delete_local_file(self, local_remote_dir: tuple[str, str]):
        local, _ = local_remote_dir
        local_file_path = os.path.join(local, 'file.txt')
        cw = CloudUploader.get(out=local)
        # Creating an empty file at specified location
        with open(local_file_path, 'w') as _:
            pass
        cw.clear_local(local_file_path)
        assert not os.path.exists(local_file_path)

    @pytest.mark.parametrize('out', ['s3://bucket/dir', 'gs://bucket/dir'])
    @pytest.mark.usefixtures('gcs_hmac_credentials')
    def test_check_bucket_exists_exception(self, out: str):
        import botocore
        with pytest.raises(botocore.exceptions.ClientError):
            _ = CloudUploader.get(out=out)

    @patch('streaming.base.storage.LocalUploader.list_objects')
    @pytest.mark.usefixtures('remote_local_dir')
    def test_list_objects_from_local_gets_called(self, mocked_requests: Mock,
                                                 remote_local_dir: Any):
        mock_remote_dir, _ = remote_local_dir()
        cu = CloudUploader.get(mock_remote_dir, exist_ok=True, keep_local=True)
        cu.list_objects()
        mocked_requests.assert_called_once()


class TestS3Uploader:

    @patch('streaming.base.storage.upload.S3Uploader.check_bucket_exists')
    @pytest.mark.parametrize('out', ['s3://bucket/dir', ('./dir1', 's3://bucket/dir/')])
    def test_instantiation(self, mocked_requests: Mock, out: Any):
        mocked_requests.side_effect = None
        _ = S3Uploader(out=out)
        if not isinstance(out, str):
            shutil.rmtree(out[0], ignore_errors=True)

    @pytest.mark.parametrize('out', ['ss4://bucket/dir'])
    def test_invalid_remote_str(self, out: str):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = S3Uploader(out=out)

    @pytest.mark.parametrize('out', ['ss4://bucket/dir', ('./dir1', 'gcs://bucket/dir/')])
    def test_invalid_remote_list(self, out: Any):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = S3Uploader(out=out)

    def test_local_directory_is_empty(self, local_remote_dir: tuple[str, str]):
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = S3Uploader(out=local)

    @pytest.mark.usefixtures('s3_client', 's3_test')
    def test_upload_file(self, local_remote_dir: tuple[str, str]):
        with tempfile.NamedTemporaryFile(delete=True, suffix='.txt') as tmp:
            filename = tmp.name.split(os.sep)[-1]
            local, _ = local_remote_dir
            remote = 's3://streaming-test-bucket/path'
            local_file_path = os.path.join(local, filename)
            s3w = S3Uploader(out=(local, remote))
            with open(local_file_path, 'w') as _:
                pass
            s3w.upload_file(filename)
            assert not os.path.exists(local_file_path)

    @pytest.mark.usefixtures('r2_client', 'r2_test')
    def test_upload_file_to_r2(self, local_remote_dir: tuple[str, str]):
        with tempfile.NamedTemporaryFile(delete=True, suffix='.txt') as tmp:
            filename = tmp.name.split(os.sep)[-1]
            local, _ = local_remote_dir
            remote = 's3://streaming-test-bucket/path'
            local_file_path = os.path.join(local, filename)
            s3w = S3Uploader(out=(local, remote))
            with open(local_file_path, 'w') as _:
                pass
            s3w.upload_file(filename)
            assert os.environ['S3_ENDPOINT_URL'] == R2_URL
            assert not os.path.exists(local_file_path)

    @pytest.mark.parametrize('out', ['s3://bucket/dir'])
    def test_check_bucket_exists_exception(self, out: str):
        import botocore

        with pytest.raises(botocore.exceptions.ClientError):
            _ = S3Uploader(out=out)

    @pytest.mark.usefixtures('s3_client', 's3_test', 'remote_local_dir')
    def test_list_objects_from_s3(self, remote_local_dir: Any):
        with tempfile.NamedTemporaryFile(delete=True, suffix='.txt') as tmp:
            file_name = tmp.name.split(os.sep)[-1]
            mock_remote_dir, _ = remote_local_dir(cloud_prefix='s3://')
            client = boto3.client('s3', region_name='us-east-1')
            client.put_object(Bucket=MY_BUCKET, Key=os.path.join(MY_PREFIX, file_name), Body='')

            cu = CloudUploader.get(mock_remote_dir, exist_ok=True, keep_local=True)
            objs = cu.list_objects(mock_remote_dir)
            assert isinstance(objs, list)

    @pytest.mark.usefixtures('s3_client', 's3_test', 'remote_local_dir')
    def test_clienterror_exception(self, remote_local_dir: Any):
        mock_remote_dir, _ = remote_local_dir(cloud_prefix='s3://')
        cu = CloudUploader.get(mock_remote_dir, exist_ok=True, keep_local=True)
        objs = cu.list_objects()
        if objs:
            assert (len(objs) == 0)

    @pytest.mark.usefixtures('s3_client', 's3_test', 'remote_local_dir')
    def test_invalid_cloud_prefix(self, remote_local_dir: Any):
        with pytest.raises(ValueError):
            mock_remote_dir, _ = remote_local_dir(cloud_prefix='s9://')
            cu = CloudUploader.get(mock_remote_dir, exist_ok=True, keep_local=True)
            _ = cu.list_objects()

    @pytest.mark.usefixtures('s3_client', 's3_test')
    def test_extra_args(self, local_remote_dir: tuple[str, str]):
        with tempfile.NamedTemporaryFile(delete=True, suffix='.txt') as tmp:
            filename = tmp.name.split(os.sep)[-1]
            local, _ = local_remote_dir
            remote = 's3://streaming-test-bucket/path'
            local_file_path = os.path.join(local, filename)
            s3w = S3Uploader(out=(local, remote))
            s3w.s3.upload_file = MagicMock()
            with mock.patch.dict('os.environ'):
                # No ACL
                with open(local_file_path, 'w') as _:
                    pass
                os.environ.pop('S3_CANNED_ACL', None)
                s3w.upload_file(filename)
                s3w.s3.upload_file.assert_called_once()
                s3w.s3.upload_file.assert_called_with(ANY, ANY, ANY, ExtraArgs={}, Callback=ANY)

                # With `authenticated-read` ACL
                with open(local_file_path, 'w') as _:
                    pass
                os.environ['S3_CANNED_ACL'] = 'authenticated-read'
                s3w.upload_file(filename)
                s3w.s3.upload_file.assert_called_with(ANY,
                                                      ANY,
                                                      ANY,
                                                      ExtraArgs={'ACL': 'authenticated-read'},
                                                      Callback=ANY)

                # With `bucket-owner-full-control` ACL
                with open(local_file_path, 'w') as _:
                    pass
                os.environ['S3_CANNED_ACL'] = 'bucket-owner-full-control'
                s3w.upload_file(filename)
                s3w.s3.upload_file.assert_called_with(
                    ANY, ANY, ANY, ExtraArgs={'ACL': 'bucket-owner-full-control'}, Callback=ANY)


class TestGCSUploader:

    @patch('streaming.base.storage.upload.GCSUploader.check_bucket_exists')
    @pytest.mark.parametrize('out', ['gs://bucket/dir', ('./dir1', 'gs://bucket/dir/')])
    @pytest.mark.usefixtures('gcs_hmac_credentials')
    def test_instantiation(self, mocked_requests: Mock, out: Any):
        mocked_requests.side_effect = None
        _ = GCSUploader(out=out)
        if not isinstance(out, str):
            shutil.rmtree(out[0], ignore_errors=True)

    @pytest.mark.parametrize('out', ['gcs://bucket/dir'])
    @pytest.mark.usefixtures('gcs_hmac_credentials')
    def test_invalid_remote_str(self, out: str):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = GCSUploader(out=out)

    @pytest.mark.parametrize('out', ['gcs://bucket/dir', ('./dir1', 'ocix://bucket/dir/')])
    @pytest.mark.usefixtures('gcs_hmac_credentials')
    def test_invalid_remote_list(self, out: Any):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = GCSUploader(out=out)

    @pytest.mark.usefixtures('gcs_hmac_credentials')
    def test_local_directory_is_empty(self, local_remote_dir: tuple[str, str]):
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = GCSUploader(out=local)

    @pytest.mark.usefixtures('gcs_hmac_client', 'gcs_test')
    def test_upload_file(self, local_remote_dir: tuple[str, str]):
        with tempfile.NamedTemporaryFile(delete=True, suffix='.txt') as tmp:
            filename = tmp.name.split(os.sep)[-1]
            local, _ = local_remote_dir
            remote = 'gs://streaming-test-bucket/path'
            local_file_path = os.path.join(local, filename)
            gcsw = GCSUploader(out=(local, remote))
            with open(local_file_path, 'w') as _:
                pass
            gcsw.upload_file(filename)
            assert not os.path.exists(local_file_path)

    @pytest.mark.usefixtures('gcs_hmac_credentials')
    @pytest.mark.parametrize('out', ['gs://bucket/dir'])
    def test_check_bucket_exists_exception(self, out: str):
        import botocore

        with pytest.raises(botocore.exceptions.ClientError):
            _ = GCSUploader(out=out)

    @patch('streaming.base.storage.upload.GCSUploader.check_bucket_exists')
    @pytest.mark.usefixtures('gcs_hmac_credentials')
    @pytest.mark.parametrize('out', ['gs://bucket/dir'])
    def test_hmac_authentication(self, mocked_requests: Mock, out: str):
        uploader = GCSUploader(out=out)
        assert uploader.authentication == GCSAuthentication.HMAC

    @patch('google.auth.default')
    @patch('google.cloud.storage.Client')
    @pytest.mark.usefixtures('gcs_service_account_credentials')
    @pytest.mark.parametrize('out', ['gs://bucket/dir'])
    def test_service_account_authentication(self, mock_client: Mock, mock_default: Mock, out: str):
        mock_default.return_value = Mock(), None
        uploader = GCSUploader(out=out)
        assert uploader.authentication == GCSAuthentication.SERVICE_ACCOUNT

    @patch('streaming.base.storage.upload.GCSUploader.check_bucket_exists')
    @patch('google.auth.default')
    @patch('google.cloud.storage.Client')
    @pytest.mark.usefixtures('gcs_service_account_credentials', 'gcs_hmac_credentials')
    @pytest.mark.parametrize('out', ['gs://bucket/dir'])
    def test_service_account_and_hmac_authentication(self, mocked_requests: Mock,
                                                     mock_default: Mock, mock_client: Mock,
                                                     out: str):
        mock_default.return_value = Mock(), None
        uploader = GCSUploader(out=out)
        assert uploader.authentication == GCSAuthentication.HMAC

    @pytest.mark.parametrize('out', ['gs://bucket/dir'])
    def test_no_authentication(self, out: str):
        with pytest.raises(
                ValueError,
                match=
            (f'Either set the environment variables `GCS_KEY` and `GCS_SECRET` or use any of '
             f'the methods in https://cloud.google.com/docs/authentication/external/set-up-adc '
             f'to set up Application Default Credentials. See also '
             f'https://docs.mosaicml.com/projects/mcli/en/latest/resources/secrets/gcp.html.')):
            _ = GCSUploader(out=out)

    @pytest.mark.usefixtures('gcs_hmac_client', 'gcs_test', 'remote_local_dir')
    def test_invalid_cloud_prefix(self, remote_local_dir: Any):
        with pytest.raises(ValueError):
            mock_remote_dir, _ = remote_local_dir(cloud_prefix='gs9://')
            cu = CloudUploader.get(mock_remote_dir, exist_ok=True, keep_local=True)
            _ = cu.list_objects()

    def test_no_credentials_error(self, remote_local_dir: Any):
        """Ensure we raise a value error correctly if we have no credentials available."""
        with pytest.raises(ValueError):
            mock_remote_dir, _ = remote_local_dir(cloud_prefix='gs://')
            cu = CloudUploader.get(mock_remote_dir, exist_ok=True, keep_local=True)
            _ = cu.list_objects()


class TestAzureUploader:

    @patch('streaming.base.storage.upload.AzureUploader.check_bucket_exists')
    @pytest.mark.usefixtures('azure_credentials')
    @pytest.mark.parametrize('out', ['azure://bucket/dir', ('./dir1', 'azure://bucket/dir/')])
    def test_instantiation(self, mocked_requests: Mock, out: Any):
        mocked_requests.side_effect = None
        _ = AzureUploader(out=out)
        if not isinstance(out, str):
            shutil.rmtree(out[0], ignore_errors=True)

    @pytest.mark.parametrize('out', ['ss4://bucket/dir'])
    def test_invalid_remote_str(self, out: str):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = AzureUploader(out=out)

    @pytest.mark.parametrize('out', ['ss4://bucket/dir', ('./dir1', 'gcs://bucket/dir/')])
    def test_invalid_remote_list(self, out: Any):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = AzureUploader(out=out)

    def test_local_directory_is_empty(self, local_remote_dir: tuple[str, str]):
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = AzureUploader(out=local)


class TestAzureDataLakeUploader:

    @patch('streaming.base.storage.upload.AzureDataLakeUploader.check_container_exists')
    @pytest.mark.usefixtures('azure_credentials')
    @pytest.mark.parametrize('out',
                             ['azure://container/dir', ('./dir1', 'azure://container/dir/')])
    def test_instantiation(self, mocked_requests: Mock, out: Any):
        mocked_requests.side_effect = None
        _ = AzureDataLakeUploader(out=out)
        if not isinstance(out, str):
            shutil.rmtree(out[0], ignore_errors=True)

    @pytest.mark.parametrize('out', ['ss4://container/dir'])
    def test_invalid_remote_str(self, out: str):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = AzureDataLakeUploader(out=out)

    @pytest.mark.parametrize('out', ['ss4://container/dir', ('./dir1', 'gcs://container/dir/')])
    def test_invalid_remote_list(self, out: Any):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = AzureDataLakeUploader(out=out)

    def test_local_directory_is_empty(self, local_remote_dir: tuple[str, str]):
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = AzureDataLakeUploader(out=local)


class TestHFUploader:

    @patch('streaming.base.storage.upload.HFUploader.check_dataset_exists')
    @pytest.mark.usefixtures('hf_credentials')
    @pytest.mark.parametrize('out', ['hf://datasets/org_name/repo_name/path'])
    def test_instantiation(self, mocked_requests: Mock, out: Any):
        mocked_requests.side_effect = None
        _ = HFUploader(out=out)
        if not isinstance(out, str):
            shutil.rmtree(out[0], ignore_errors=True)

    @pytest.mark.parametrize('out', ['ss4://container/dir'])
    def test_invalid_remote_str(self, out: str):
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = HFUploader(out=out)

    def test_local_directory_is_empty(self, local_remote_dir: tuple[str, str]):
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = HFUploader(out=local)


class TestDatabricksUnityCatalogUploader:

    @patch('streaming.base.storage.upload.DatabricksUploader._create_workspace_client')
    @pytest.mark.parametrize(
        'out', ['dbfs:/Volumes/container/dir', ('./dir1', 'dbfs:/Volumes/container/dir/')])
    def test_instantiation(self, mock_create_client: Mock, out: Any):
        mock_create_client.side_effect = None
        _ = DatabricksUnityCatalogUploader(out=out)
        if not isinstance(out, str):
            shutil.rmtree(out[0], ignore_errors=True)

    @patch('streaming.base.storage.upload.DatabricksUploader._create_workspace_client')
    @pytest.mark.parametrize('out', ['ss4://bucket/dir', ('./dir1', 'gcs://bucket/dir/')])
    def test_invalid_remote_list(self, mock_create_client: Mock, out: Any):
        mock_create_client.side_effect = None
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = DatabricksUnityCatalogUploader(out=out)

    @patch('streaming.base.storage.upload.DatabricksUploader._create_workspace_client')
    def test_local_directory_is_empty(self, mock_create_client: Mock,
                                      local_remote_dir: tuple[str, str]):
        mock_create_client.side_effect = None
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = DatabricksUnityCatalogUploader(out=local)


class TestDBFSUploader:

    @patch('streaming.base.storage.upload.DatabricksUploader._create_workspace_client')
    @pytest.mark.parametrize('out', ['dbfs:/container/dir', ('./dir1', 'dbfs:/container/dir/')])
    def test_instantiation(self, mock_create_client: Mock, out: Any):
        mock_create_client.side_effect = None
        _ = DBFSUploader(out=out)
        if not isinstance(out, str):
            shutil.rmtree(out[0], ignore_errors=True)

    @patch('streaming.base.storage.upload.DatabricksUploader._create_workspace_client')
    @pytest.mark.parametrize('out', ['ss4://bucket/dir', ('./dir1', 'gcs://bucket/dir/')])
    def test_invalid_remote_list(self, mock_create_client: Mock, out: Any):
        mock_create_client.side_effect = None
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = DBFSUploader(out=out)

    @patch('streaming.base.storage.upload.DatabricksUploader._create_workspace_client')
    def test_local_directory_is_empty(self, mock_create_client: Mock,
                                      local_remote_dir: tuple[str, str]):
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            mock_create_client.side_effect = None
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = DBFSUploader(out=local)


class TestAlipanUploader:

    @patch('streaming.base.storage.upload.AlipanUploader.check_token')
    @pytest.mark.usefixtures('alipan_credentials')
    @pytest.mark.parametrize('out',
                             ['alipan:///container/dir', ('./dir1', 'alipan:///container/dir/')])
    def test_instantiation(self, mock_create_client: Mock, out: Any):
        mock_create_client.side_effect = None
        _ = AlipanUploader(out=out)
        if not isinstance(out, str):
            shutil.rmtree(out[0], ignore_errors=True)

    @patch('streaming.base.storage.upload.AlipanUploader.check_token')
    @pytest.mark.usefixtures('alipan_credentials')
    @pytest.mark.parametrize('out', ['alipann://bucket/dir', ('./dir1', 'gcs://bucket/dir/')])
    def test_invalid_remote_list(self, mock_create_client: Mock, out: Any):
        mock_create_client.side_effect = None
        with pytest.raises(ValueError, match=f'Invalid Cloud provider prefix.*'):
            _ = AlipanUploader(out=out)

    @patch('streaming.base.storage.upload.AlipanUploader.check_token')
    @pytest.mark.usefixtures('alipan_credentials')
    def test_local_directory_is_empty(self, mock_create_client: Mock,
                                      local_remote_dir: tuple[str, str]):
        with pytest.raises(FileExistsError, match=f'Directory is not empty.*'):
            mock_create_client.side_effect = None
            local, _ = local_remote_dir
            os.makedirs(local, exist_ok=True)
            local_file_path = os.path.join(local, 'file.txt')
            # Creating an empty file at specified location
            with open(local_file_path, 'w') as _:
                pass
            _ = AlipanUploader(out=local)


class TestLocalUploader:

    def test_upload_file(self, local_remote_dir: tuple[str, str]):
        local, remote = local_remote_dir
        filename = 'file.txt'
        local_file_path = os.path.join(local, filename)
        remote_file_path = os.path.join(remote, filename)
        lw = LocalUploader(out=(local, remote))
        # Creating an empty file at specified location
        with open(local_file_path, 'w') as _:
            pass
        assert not os.path.exists(remote_file_path)
        lw.upload_file(filename)
        assert os.path.exists(remote_file_path)

    def test_instantiation_remote_none(self, local_remote_dir: tuple[str, str]):
        local, _ = local_remote_dir
        lc = LocalUploader(out=local)
        assert lc.local == local
        assert lc.remote is None

    def test_upload_file_remote_none(self, local_remote_dir: tuple[str, str]):
        local, remote = local_remote_dir
        filename = 'file.txt'
        local_file_path = os.path.join(local, filename)
        remote_file_path = os.path.join(remote, filename)
        lc = LocalUploader(out=local)
        # Creating an empty file at specified location
        with open(local_file_path, 'w') as _:
            pass
        lc.upload_file(filename)
        assert not os.path.exists(remote_file_path)

    def test_upload_file_from_local_to_remote(self, local_remote_dir: tuple[str, str]):
        local, remote = local_remote_dir
        filename = 'file.txt'
        local_file_path = os.path.join(local, filename)
        remote_file_path = os.path.join(remote, filename)
        lc = LocalUploader(out=(local, remote))
        # Creating an empty file at specified location
        with open(local_file_path, 'w') as _:
            pass
        lc.upload_file(filename)
        assert os.path.exists(remote_file_path)

    def test_upload_file_exception(self, local_remote_dir: tuple[str, str]):
        local, remote = local_remote_dir
        filename = 'file.txt'
        lc = LocalUploader(out=(local, remote))
        with pytest.raises(FileNotFoundError, match=f'No such file or directory:.*'):
            lc.upload_file(filename)

    def test_list_objects_no_prefix(self, local_remote_dir: tuple[str, str]):
        local, _ = local_remote_dir
        lc = LocalUploader(out=local)

        # Generate some local files for testing
        test_dir = os.path.join(local, 'test_dir')
        os.makedirs(test_dir, exist_ok=True)
        with open(os.path.join(test_dir, 'file2.txt'),
                  'w') as f2, open(os.path.join(test_dir, 'file1.txt'), 'w') as f1:
            f1.write('Content of file1')
            f2.write('Content of file2')

        result = lc.list_objects()
        expected = [os.path.join(test_dir, 'file1.txt'), os.path.join(test_dir, 'file2.txt')]
        assert (result == expected)

    def test_list_objects_with_prefix(self, local_remote_dir: tuple[str, str]):
        local, _ = local_remote_dir
        lc = LocalUploader(out=local)

        # Generate some local files for testing
        test_dir = os.path.join(local, 'test_dir')
        os.makedirs(test_dir, exist_ok=True)
        with open(os.path.join(test_dir, 'prefix_file2.txt'),
                  'w') as f2, open(os.path.join(test_dir, 'prefix_file1.txt'), 'w') as f1:
            f1.write('Content of file1')
            f2.write('Content of file2')

        result = lc.list_objects(prefix='test_dir/prefix')
        expected = [
            os.path.join(test_dir, 'prefix_file1.txt'),
            os.path.join(test_dir, 'prefix_file2.txt')
        ]
        assert (result == expected)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Functions for simulating shard downloads and calculating needed cache limit for downloads."""

from typing import Optional

import numpy as np
from core.node_tracker import NodeTracker
from numpy.typing import NDArray


def simulate_shard_downloads(node: NodeTracker,
                             raw_shard_sizes: NDArray[np.int64],
                             zip_shard_sizes: NDArray[np.int64],
                             current_batch_downloads: bool,
                             step_num: int,
                             cache_limit: Optional[int] = None,
                             shards_needed: Optional[set] = None,
                             download_bytes_left: Optional[int] = None) -> tuple[str, int]:
    """Simulate downloading a shard for a node.

    Args:
        node (NodeTracker): The node to simulate downloading a shard for.
        raw_shard_sizes (NDArray[np.int64]): The raw sizes of all shards.
        zip_shard_sizes (NDArray[np.int64]): The zip sizes of all shards.
        current_batch_downloads (bool): Whether we are downloading shards for the current batch.
        step_num (int): The current step number.
        cache_limit (Optional[int]): The cache limit for the node. Defaults to ``None``.
        shards_needed (Optional[set]): The shards needed for the current batch.
            Defaults to ``None``.
        download_bytes_left (Optional[int]): The number of download bytes left in the downloading
            time interval. Defaults to ``None``.

    Returns:
        Tuple[bool, int]: A tuple of the shard download status and the download size.
    """
    worker_download = node.get_next_worker_with_downloads()
    if worker_download is None:
        # No workers have shards to download.
        return ('empty', 0)

    # Proceed with downloading the shard for this worker.
    download_shard = int(worker_download[0])

    # Get the raw and zip sizes, in bytes, of the shard.
    shard_raw_size = int(raw_shard_sizes[download_shard])
    shard_zip_size = int(zip_shard_sizes[download_shard])
    # If shard is compressed, we download the zipped size. Otherwise, download raw size.
    download_size = shard_zip_size or shard_raw_size

    # If we are not downloading for the current batch, we need to check if the download bytes
    # left is sufficient to download this shard. Otherwise, we have to keep downloading anyways.
    bytes_sufficient = True
    if not current_batch_downloads:
        if download_bytes_left is not None:
            bytes_sufficient = (download_size <= download_bytes_left)
        else:
            raise ValueError('Must specify download_bytes_left if not downloading for \
                             current batch.')

    if download_shard not in node.shards and bytes_sufficient:
        # Shard is not present in node, so we download it.
        # Handle possible shard eviction.
        if cache_limit and node.cache_usage + shard_raw_size > cache_limit:
            # Evict shards until we have space for this shard.
            node.evict_until_satisfied(shard_raw_size, raw_shard_sizes)

        # Shards are assumed to be decompressed on download, so cache_usage increases by raw size.
        node.cache_usage += shard_raw_size

        # If we are downloading shards for the current batch, we need to check if the shard
        # is needed by the current batch. If it is, we make sure to mark it as most recently used.
        # If we are not downloading shards for the current batch then no shard is marked as used.
        if current_batch_downloads:
            if shards_needed is not None:
                node.add_shard(download_shard)
                shards_needed.discard(download_shard)
            else:
                raise ValueError('Must specify shards_needed if downloading for current batch.')
        else:
            node.add_shard(download_shard)

        if node.shard_access_starts[download_shard] == -1:
            # Shard has never been accessed before. Set its access start.
            node.shard_access_starts[download_shard] = step_num
        # For any shard access, we are accessing the shard so we need the shard until
        # at least the next step begins. Adding 0.5 ensures that we evict shards
        # after they are used for the last time, but before they are replaced by
        # new downloads in the next step.
        node.shard_access_ends[download_shard] = step_num + 0.5

        # Advance the worker download index.
        node.increment_worker_download_index()
        # We have now downloaded this shard. Remove from worker download queue.
        worker_download.pop()
        return ('downloaded', download_size)
    elif not (current_batch_downloads) and download_shard not in node.shards \
        and download_bytes_left is not None:
        # This is the case when we are not downloading for the current batch, and need to download
        # a shard but do not have the download bytes to fully download the shard this step.
        # We do not advance the worker download index since we still are downloading this shard.
        node.partial_shard_bytes = download_bytes_left
        # We only account for downloaded bytes when we fully download a shard.
        return ('partial', 0)
    else:
        # The shard is already present in the node. Advance the worker download index.
        node.increment_worker_download_index()
        # Node already has this shard. Remove from worker download queue.
        worker_download.pop()
        return ('present', 0)


def run_cache_limit(nodes: list[NodeTracker], raw_shard_sizes: NDArray[np.int64]) -> int:
    """Find the minimum needed cache limit across all nodes for this run.

    Args:
        nodes (list[NodeTracker]): The nodes, which contain shard use information.
        raw_shard_sizes (NDArray[np.int64]): The raw sizes of all shards.

    Returns:
        int: The minimum needed cache limit, in bytes, for the run.
    """
    # Find the overall needed cache usage, as the max needed for any node at any point.
    needed_cache_usage = 0
    for node in nodes:
        # For each node, create its own list of shard access events.
        # Access event tuples are (event time, shard idx, event type)
        # Event types: 0 means a shard has been accessed, 1 means a shard has ended access.
        node_shards = node.get_all_shards()
        access_events = []
        access_events += [(node.shard_access_starts[i], i, 0) for i in node_shards]
        access_events += [(node.shard_access_ends[i], i, 1) for i in node_shards]

        # Sort the access events to get shard events, in order.
        access_events.sort(key=lambda x: x[0])

        # For each shard event, update the cache usage. Assume that shards are decompressed
        # immediately on download, so use raw shard sizes.
        curr_cache_usage = 0
        for event in access_events:
            if event[2] == 0:
                # Shard has been accessed. Increment cache usage.
                curr_cache_usage += raw_shard_sizes[event[1]]
                # Needed cache usage is the max of all cache usages across the run.
                if curr_cache_usage > needed_cache_usage:
                    needed_cache_usage = curr_cache_usage
            else:
                # Shard access has ended. Decrement cache usage.
                curr_cache_usage -= raw_shard_sizes[event[1]]

    return needed_cache_usage



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Create a dataset index file from input parameters."""

import json
import logging
import os
import random
import string
from typing import Optional

from streaming.base.format import get_index_basename

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def get_random_foldername() -> str:
    """Generate random folder name to store the index file in.

    Returns:
        str: random alphanumeric folder name.
    """
    return ''.join(
        random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits)
        for _ in range(16))


def create_stream_index(shards: int, samples_per_shard: int, avg_raw_shard_size: int,
                        avg_zip_shard_size: Optional[int]) -> str:
    """Create dataset index file from input parameters.

    Args:
        shards (int): Number of shards.
        samples_per_shard (int): Number of samples per shard.
        avg_raw_shard_size (int): Average raw shard size.
        avg_zip_shard_size (int): Average compressed shard size.

    Returns:
        local path to created index file for stream.
    """
    index_data = {'version': 2, 'shards': []}

    shards_list = []
    for _ in range(shards):
        shard_data = {
            'column_encodings': [],
            'column_names': [],
            'column_sizes': [],
            'format': 'mds',
            'raw_data': {
                'basename': '',
                'bytes': avg_raw_shard_size,
                'hashes': {}
            },
            'hashes': [],
            'samples': samples_per_shard,
            'size_limit': avg_raw_shard_size,
            'version': 2,
            'zip_data': None,
            'compression': None
        }
        if avg_zip_shard_size is not None:
            shard_data['zip_data'] = {'basename': '', 'bytes': avg_zip_shard_size, 'hashes': {}}
            shard_data['compression'] = ''
        shards_list.append(shard_data)

    index_data['shards'] = shards_list

    # Try making the directory for the stream's index.json file
    foldername = get_random_foldername() + '_indexcreated'
    try:
        os.mkdir(foldername)
    except FileExistsError:
        logger.warning(' Folder already exists, trying again...')
        foldername = get_random_foldername()
        os.mkdir(foldername)

    index_basename = get_index_basename()
    index_path = os.path.join(foldername, index_basename)

    with open(index_path, 'w') as f:
        json.dump(index_data, f)

    return index_path



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Apportion shards/samples such that batches have samples randomly selected from streams."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray

from streaming.base.partition import get_partitions
from streaming.base.shuffle import get_shuffle
from streaming.base.world import World

if TYPE_CHECKING:
    from streaming.base.dataset import StreamingDataset

logger = logging.getLogger(__name__)


def generate_work_random_batching(dataset: StreamingDataset, world: World, epoch: int,
                                  sample_in_epoch: int) -> NDArray[np.int64]:
    """Generate this epoch's arrangement of samples for ``random`` batching.

    This is only called in local rank zero. When ``batching_method`` is set to ``per_stream``,
    which is the default case, each batch consists of samples selected at random from across
    all streams.

    Args:
        dataset (StreamingDataset): Dataset to generate the partition for.
        world (World): World state.
        epoch (int): Which epoch it is.
        sample_in_epoch (int): Where we are in the epoch.

    Returns:
        NDArray[np.int64]: The epoch (num physical nodes, ranks per node, workers per rank,
            batches per worker, batch size).
    """
    # Ensure that num_canonical_nodes has been set.
    if dataset.num_canonical_nodes is None:
        raise RuntimeError(f'`num_canonical_nodes` can never be None. ' +
                           f'Provide a positive integer.')

    # Sample each shard of each stream according to their proportions/repeats/samples. This
    # gives us the resampled size of each underlying shard, and a mapping from each fake "big"
    # sample ID to its underlying "small" sample ID.
    shuffle_units, small_per_big = dataset.resample_streams(epoch)

    batch_size = dataset.batch_size
    assert isinstance(batch_size, int), f'Batch size must be an integer. Got {type(batch_size)}.'

    # Partition the global sample space (of resampled "big" sample IDs) into a tensor of shape
    # (num physical nodes, ranks per node, workers per rank, batches per worker, samples per
    # batch) such that we have an elastically deterministic sample order.
    big_ids = get_partitions(dataset.partition_algo, dataset.epoch_size,
                             dataset.num_canonical_nodes, world.num_nodes, world.ranks_per_node,
                             world.workers_per_rank, batch_size, sample_in_epoch,
                             dataset.initial_physical_nodes)

    # If we need to shuffle, shuffle in a node-aware and *underlying* shard-aware way.
    if dataset.shuffle:
        if not isinstance(dataset.shuffle_block_size, int):
            raise TypeError(f'Dataset `shuffle_block_size` must be an integer. ' +
                            f'Got {type(dataset.shuffle_block_size)} instead.')
        shuffle = get_shuffle(dataset.shuffle_algo, shuffle_units, dataset.num_canonical_nodes,
                              dataset.shuffle_seed, epoch, dataset.shuffle_block_size)
        big_ids = np.where(big_ids != -1, shuffle[big_ids], -1)

    # Now that we have partitioning and shuffled with hallucinated "big" sample IDs, we don't
    # need them anymore, and can convert back to underlying "small" sample IDs.
    return np.where(big_ids != -1, small_per_big[big_ids], -1)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""A non-streaming pytorch map Dataset."""

import json
import os
from typing import Any, Optional

import numpy as np
from torch.utils.data import Dataset

from streaming.base.array import Array
from streaming.base.format import get_index_basename, reader_from_json
from streaming.base.spanner import Spanner

__all__ = ['LocalDataset']


class LocalDataset(Array, Dataset):
    """A streaming dataset whose shards reside locally as a pytorch Dataset.

    Args:
        local (str): Local dataset directory where shards are cached by split.
        split (str, optional): Which dataset split to use, if any. Defaults to ``None``.
    """

    def __init__(self, local: str, split: Optional[str] = None):
        split = split or ''

        self.local = local
        self.split = split

        filename = os.path.join(local, split, get_index_basename())  # pyright: ignore
        obj = json.load(open(filename))
        if obj['version'] != 2:
            raise ValueError(f'Unsupported streaming data version: {obj["version"]}. ' +
                             f'Expected version 2.')

        self.shards = []
        for info in obj['shards']:
            shard = reader_from_json(local, split, info)
            self.shards.append(shard)
        self.num_samples = sum([shard.samples for shard in self.shards])

        shard_sizes = np.array([x.samples for x in self.shards])
        self.spanner = Spanner(shard_sizes)

    def __len__(self) -> int:
        """Get the length as a PyTorch Dataset.

        Returns:
            int: Dataset length.
        """
        return self.num_samples

    @property
    def size(self) -> int:
        """Get the size of the dataset in samples.

        Returns:
            int: Number of samples.
        """
        return self.num_samples

    def get_item(self, sample_id: int) -> dict[str, Any]:
        """Get sample by global sample ID.

        Args:
            sample_id (int): Sample ID.

        Returns:
            Dict[str, Any]: Column name with sample data.
        """
        shard_id, index_in_shard = self.spanner[sample_id]
        shard = self.shards[shard_id]
        return shard[index_in_shard]



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Ingest yaml and create SimulationDataset."""

from typing import Optional

from core.sim_dataset import SimulationDataset
from core.sim_time import Time, TimeUnit, ensure_time
from omegaconf import DictConfig
from omegaconf import OmegaConf as om

from streaming.base import Stream


def ingest_yaml(yaml_dict: Optional[dict] = None,
                filepath: Optional[str] = None) -> tuple[Optional[int], int, Time, int, dict]:
    """Create SimulationDataset from yaml file and other needed args.

    Args:
        yaml_dict (Optional[dict]): yaml file already converted to a dictionary
        filepath (Optional[str]): path to yaml file

    Returns:
        Tuple[Optional[int], Optional[int], Time, Optional[int], Optional[dict]]: total_devices,
            workers, max_duration, global_batch_size, train_dataset parameters from yaml
    """
    config = None
    # Read in the yaml file
    if filepath is not None:
        with open(filepath) as f:
            config = om.load(f)
            if not isinstance(config, DictConfig):
                raise ValueError('Yaml file must be a dictionary, not a list.')
    elif yaml_dict is not None:
        config = om.create(yaml_dict)
    else:
        raise ValueError('Must specify either filepath or yaml_dict.')

    # Get the number of devices (GPUs)
    if 'compute' in config and 'gpus' in config['compute']:
        total_devices = int(config['compute']['gpus'])
    else:
        total_devices = None

    workers = None
    train_dataset = None
    max_duration = None
    global_batch_size = None
    # Get the training and dataset params
    if 'parameters' in config:
        config = om.create(config['parameters'])

    om.resolve(config)

    if not isinstance(config, DictConfig):
        raise TypeError(f'`config` must be of type DictConfig. Got type {type(config)}.')

    # get global batch size
    if 'global_train_batch_size' in config:
        global_batch_size = config['global_train_batch_size']
    elif 'batch_size' in config:
        global_batch_size = config['batch_size']
    elif 'dataset' in config and 'train_batch_size' in config['dataset']:
        global_batch_size = config['dataset']['train_batch_size']

    # get number of workers and training dataset params
    if 'train_loader' in config:
        train_loader = config['train_loader']
        if 'num_workers' in train_loader:
            workers = train_loader['num_workers']
        else:
            workers = 1
        if 'dataset' in train_loader:
            train_dataset = train_loader['dataset']
        else:
            raise ValueError('dataset must be specified in the yaml file.')
    elif 'dataset' in config:
        dataset = config['dataset']
        if 'train_dataset' in dataset:
            train_dataset = dataset['train_dataset']
            if 'streaming_kwargs' in train_dataset:
                # Merge streaming kwargs, if present, into train_dataset
                train_dataset.update(train_dataset['streaming_kwargs'])
            if 'dataloader_kwargs' in train_dataset and 'num_workers' in train_dataset[
                    'dataloader_kwargs']:
                workers = train_dataset['dataloader_kwargs']['num_workers']
            else:
                workers = 1
        else:
            raise ValueError('train_dataset must be specified in the yaml file.')
    elif 'train_dataset' in config:
        train_dataset = config['train_dataset']
        if 'streaming_kwargs' in train_dataset:
            # Merge streaming kwargs, if present, into train_dataset
            train_dataset.update(train_dataset['streaming_kwargs'])
        if 'dataloader_kwargs' in train_dataset and 'num_workers' in train_dataset[
                'dataloader_kwargs']:
            workers = train_dataset['dataloader_kwargs']['num_workers']
        else:
            workers = 1
    else:
        raise ValueError('train_loader or dataset must be specified in the yaml file.')

    # Get duration of training from config
    if 'max_duration' in config:
        max_duration = config['max_duration']
    elif 'trainer' in config and 'max_duration' in config['trainer']:
        max_duration = config['trainer']['max_duration']
    else:
        raise ValueError('max_duration must be specified in the yaml file.')

    # convert max_duration to epochs or batches.
    max_duration = ensure_time(max_duration, TimeUnit.EPOCH)
    time_unit = max_duration.unit
    if time_unit != TimeUnit.EPOCH and time_unit != TimeUnit.BATCH:
        raise ValueError('Simulator currently only supports max_duration in epochs or batches.')

    # convert train_dataset to dict, if it isn't already
    if isinstance(train_dataset, DictConfig):
        train_dataset = om.to_container(train_dataset)

    if not isinstance(workers, int):
        raise ValueError(f'`workers` must be an int. Instead, got {type(workers)}.')
    if not isinstance(global_batch_size, int):
        raise ValueError(f'`global_batch_size` must be an int. Instead, got ' +
                         f'{type(global_batch_size)}.')
    if not isinstance(train_dataset, dict):
        raise ValueError(f'`train_dataset` must be a dict. Instead, got ' +
                         f'{type(train_dataset)}.')

    return total_devices, workers, max_duration, global_batch_size, train_dataset


def create_simulation_dataset(nodes: int, devices: int, workers: int, global_batch_size: int,
                              train_dataset: dict) -> SimulationDataset:
    """Create SimulationDataset from input information.

    Args:
        nodes (int): number of physical nodes
        devices (int): number of devices per node
        workers (int): number of workers per device
        global_batch_size (int): global batch size (samples)
        train_dataset (dict): train_dataset parameters from yaml file

    Returns:
        SimulationDataset: SimulationDataset created from input information.
    """
    streams = None
    # Check for cases where local and remote may be lists and turn those into streams.
    if 'local' in train_dataset and 'remote' in train_dataset:
        if isinstance(train_dataset['local'], list) \
            and isinstance(train_dataset['remote'], list):
            if len(train_dataset['local']) != len(train_dataset['remote']):
                raise ValueError('local and remote must be the same length in the yaml file.')
            streams = []
            for local, remote in zip(train_dataset['local'], train_dataset['remote']):
                streams.append(Stream(local=local, remote=remote, split=train_dataset['split'] \
                                      if 'split' in train_dataset else None))
            del train_dataset['local']
            del train_dataset['remote']

    # Don't re-retrieve streams if we have built it from local and remote lists.
    if not isinstance(streams, list):
        streams_dict = train_dataset.get('streams', None)
        if streams_dict is not None:
            streams = []
            if not isinstance(streams_dict, dict):
                raise TypeError(f'`streams` must be of type dict, if not a list. ' +
                                f'Got type {type(streams_dict)}.')
            for stream in streams_dict.values():
                if 'path' in stream:
                    del stream['path']
                # Create Stream object from each dictionary entry
                streams.append(Stream(**stream))

    remote = train_dataset.get('remote', None)
    local = train_dataset.get('local', None)
    split = train_dataset.get('split', None)
    download_retry = train_dataset.get('download_retry', 2)
    download_timeout = train_dataset.get('download_timeout', 60)
    validate_hash = train_dataset.get('validate_hash', None)
    keep_zip = train_dataset.get('keep_zip', False)
    epoch_size = train_dataset.get('epoch_size', None)
    predownload = train_dataset.get('predownload', None)
    cache_limit = train_dataset.get('cache_limit', None)
    partition_algo = train_dataset.get('partition_algo', 'relaxed')
    num_canonical_nodes = train_dataset.get('num_canonical_nodes', None)
    if global_batch_size % (devices * nodes) != 0:
        raise ValueError('global_batch_size must be divisible by total number of devices.')
    batch_size = global_batch_size // (devices * nodes)
    shuffle = train_dataset.get('shuffle', False)
    shuffle_algo = train_dataset.get('shuffle_algo', 'py1e')
    shuffle_seed = train_dataset.get('shuffle_seed', 9176)
    shuffle_block_size = train_dataset.get('shuffle_block_size', None)
    sampling_method = train_dataset.get('sampling_method', 'balanced')
    sampling_granularity = train_dataset.get('sampling_granularity', 1)
    batching_method = train_dataset.get('batching_method', 'random')

    dataset = SimulationDataset(nodes, devices, workers, streams, remote, local, split,
                                download_retry, download_timeout, validate_hash, keep_zip,
                                epoch_size, predownload, cache_limit, partition_algo,
                                num_canonical_nodes, batch_size, shuffle, shuffle_algo,
                                shuffle_seed, shuffle_block_size, sampling_method,
                                sampling_granularity, batching_method)

    return dataset



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Base classes for computer vision :class:`StreamingDataset`s."""

from typing import Any, Callable, Optional

from torchvision.datasets import VisionDataset
from torchvision.transforms.functional import to_tensor

from streaming.base import StreamingDataset

__all__ = ['StreamingVisionDataset']


class StandardTransform:
    """Individual input and output transforms called jointly, following torchvision.

    Args:
        transform (Callable, optional): Input transform. Defaults to ``None``.
        target_transform (Callable, optional): Output transform. Defaults to ``None``.
    """

    def __init__(self,
                 transform: Optional[Callable] = None,
                 target_transform: Optional[Callable] = None) -> None:
        self.transform = transform
        self.target_transform = target_transform

    def __call__(self, x: Any, y: Any) -> tuple[Any, Any]:
        """Apply the transforms to input and output.

        Args:
            x (Any): Input.
            y (Any): Output.

        Returns:
            Tuple[Any, Any]: Transformed input and output.
        """
        if self.transform:
            x = self.transform(x)
        else:
            x = to_tensor(x)
        if self.target_transform:
            y = self.target_transform(y)
        return x, y


class StreamingVisionDataset(StreamingDataset, VisionDataset):
    """A streaming, iterable, torchvision VisionDataset.

    Args:
        remote (str, optional): Remote path or directory to download the dataset from. If ``None``,
            its data must exist locally. StreamingDataset uses either ``streams`` or
            ``remote``/``local``. Defaults to ``None``.
        local (str, optional): Local working directory to download shards to. This is where shards
            are cached while they are being used. Uses a temp directory if not set.
            StreamingDataset uses either ``streams`` or ``remote``/``local``. Defaults to ``None``.
        split (str, optional): Which dataset split to use, if any. If provided, we stream from/to
            the ``split`` subdirs of  ``remote`` and ``local``. Defaults to ``None``.
        download_retry (int): Number of download re-attempts before giving up. Defaults to ``2``.
        download_timeout (float): Number of seconds to wait for a shard to download before raising
            an exception. Defaults to ``60``.
        validate_hash (str, optional): Optional hash or checksum algorithm to use to validate
            shards. Defaults to ``None``.
        keep_zip (bool): Whether to keep or delete the compressed form when decompressing
            downloaded shards. If ``False``, keep iff remote is local or no remote. Defaults to
            ``False``.
        epoch_size (int, optional): Number of samples to draw per epoch balanced across all
            streams. If ``None``, takes its value from the total number of underlying samples.
            Provide this field if you are weighting streams relatively to target a larger or
            smaller epoch size. Defaults to ``None``.
        predownload (int, optional): Target number of samples to download per worker in advance
            of current sample. Workers will attempt to download ahead by this many samples during,
            but not before, training. Recommendation is to provide a value greater than per device
            batch size to ensure at-least per device batch size number of samples cached locally.
            If ``None``, its value gets derived using per device batch size and number of
            canonical nodes ``max(batch_size, 256 * batch_size // num_canonical_nodes)``.
            Defaults to ``None``.
        cache_limit (int, optional): Maximum size in bytes of this StreamingDataset's shard cache.
            Before downloading a shard, the least recently used resident shard(s) may be evicted
            (deleted from the local cache) in order to stay under the limit. Set to ``None`` to
            disable shard eviction. Defaults to ``None``.
        partition_algo (str): Which partitioning algorithm to use. Defaults to ``orig``.
        num_canonical_nodes (int, optional): Canonical number of nodes for shuffling with
            resumption. The sample space is divided evenly according to the number of canonical
            nodes. The higher the value, the more independent non-overlapping paths the
            StreamingDataset replicas take through the shards per model replica (increasing data
            source diversity). Defaults to ``None``, which is interpreted as 64 times the number
            of nodes of the initial run.

            .. note::

                For sequential sample ordering, set ``shuffle`` to ``False`` and
                ``num_canonical_nodes`` to the number of physical nodes of the initial run.
        batch_size (int, optional): Batch size of its DataLoader, which affects how the dataset is
            partitioned over the workers. Defaults to ``None``.
        shuffle (bool): Whether to iterate over the samples in randomized order. Defaults to
            ``False``.
        shuffle_algo (str): Which shuffling algorithm to use. Defaults to ``py1s``.
        shuffle_seed (int): Seed for Deterministic data shuffling. Defaults to ``9176``.
        shuffle_block_size (int): Unit of shuffle. Defaults to ``1 << 18``.
        transforms (callable, optional): A function/transforms that takes in an image and a label
            and returns the transformed versions of both. Defaults to ``None``.
        transform (callable, optional): A function/transform that takes in an image and returns a
            transformed version. Defaults to ``None``.
        target_transform (callable, optional): A function/transform that takes in a target and
            returns a transformed version. Defaults to ``None``.
    """

    def __init__(self,
                 *,
                 remote: Optional[str] = None,
                 local: Optional[str] = None,
                 split: Optional[str] = None,
                 download_retry: int = 2,
                 download_timeout: float = 60,
                 validate_hash: Optional[str] = None,
                 keep_zip: bool = False,
                 epoch_size: Optional[int] = None,
                 predownload: Optional[int] = None,
                 cache_limit: Optional[int] = None,
                 partition_algo: str = 'orig',
                 num_canonical_nodes: Optional[int] = None,
                 batch_size: int,
                 shuffle: bool = False,
                 shuffle_algo: str = 'py1s',
                 shuffle_seed: int = 9176,
                 shuffle_block_size: int = 1 << 18,
                 transforms: Optional[Callable] = None,
                 transform: Optional[Callable] = None,
                 target_transform: Optional[Callable] = None) -> None:
        StreamingDataset.__init__(self,
                                  remote=remote,
                                  local=local,
                                  split=split,
                                  download_retry=download_retry,
                                  download_timeout=download_timeout,
                                  validate_hash=validate_hash,
                                  keep_zip=keep_zip,
                                  epoch_size=epoch_size,
                                  predownload=predownload,
                                  cache_limit=cache_limit,
                                  partition_algo=partition_algo,
                                  num_canonical_nodes=num_canonical_nodes,
                                  batch_size=batch_size,
                                  shuffle=shuffle,
                                  shuffle_algo=shuffle_algo,
                                  shuffle_seed=shuffle_seed,
                                  shuffle_block_size=shuffle_block_size)

        has_transforms = transforms is not None
        has_separate_transform = transform is not None or target_transform is not None
        if has_transforms and has_separate_transform:
            raise ValueError(
                'Only transforms or transform/target_transform can be passed as an argument')

        self.transform = transform
        self.target_transform = target_transform
        if not has_transforms:
            transforms = StandardTransform(transform, target_transform)
        self.transforms = transforms

    def get_item(self, idx: int) -> Any:
        """Get sample by global index, blocking to load its shard if missing.

        Args:
            idx (int): Sample index.

        Returns:
            Any: Sample data.
        """
        obj = super().get_item(idx)
        x = obj['x']
        y = obj['y']
        return self.transforms(x, y)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Mapping of global sample index to shard and relative sample index."""

import numpy as np
from numpy.typing import NDArray


class Spanner:
    """Given a list of shards, construct a mapping of global index to shard and relative index.

    Args:
        shard_sizes (NDArray[np.int64]): Number of samples in each shard.
        span_size (int): Size of the divisions of the sample space. Defaults to ``1 << 10``.
    """

    def __init__(self, shard_sizes: NDArray[np.int64], span_size: int = 1 << 10) -> None:
        self.shard_sizes = shard_sizes
        self.span_size = span_size
        self.num_samples = sum(shard_sizes)
        self.shard_bounds = np.concatenate([np.zeros(1, np.int64), shard_sizes.cumsum()])

        overflow = self.num_samples % span_size
        underflow = span_size - overflow if overflow else 0
        self.shard_sizes[-1] += underflow

        sample_shards = np.repeat(np.arange(len(shard_sizes)), self.shard_sizes)
        sample_shards = sample_shards.reshape(-1, span_size)
        span_lowest_shards = sample_shards.min(1)
        span_highest_shards = sample_shards.max(1)

        self.spans = []
        for low, high in zip(span_lowest_shards, span_highest_shards):
            shards = np.arange(low, high + 1)
            self.spans.append(shards)

        self.shard_sizes[-1] -= underflow

    def __getitem__(self, index: int) -> tuple[int, int]:
        """Map global sample index to shard and relative sample index.

        Args:
            index (int): Global sample index.

        Returns:
            Tuple[int, int]: Shard and relative sample index.
        """
        if not (0 <= index < self.num_samples):
            raise IndexError(f'Invalid sample index `{index}`: 0 <= {index} < {self.num_samples}')

        span = index // self.span_size
        for shard in self.spans[span]:
            shard_start = self.shard_bounds[shard]
            shard_stop = self.shard_bounds[shard + 1]
            if shard_start <= index < shard_stop:
                return shard, int(index - shard_start.item())  # pyright: ignore

        raise RuntimeError('Internal error: shards were indexed incorrectly')



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import pytest

from streaming import MDSWriter, StreamingDataset


@pytest.mark.usefixtures('local_remote_dir')
def test_do_allow_unsafe_types_safe(local_remote_dir: tuple[str, str]):
    local, _ = local_remote_dir
    columns = {'num': 'int'}
    with MDSWriter(out=local, columns=columns) as out:
        for num in range(100):
            sample = {'num': num}
            out.write(sample)
    StreamingDataset(local=local, allow_unsafe_types=True, batch_size=1)


@pytest.mark.usefixtures('local_remote_dir')
def test_do_allow_unsafe_types_unsafe(local_remote_dir: tuple[str, str]):
    local, _ = local_remote_dir
    columns = {'num': 'pkl'}
    with MDSWriter(out=local, columns=columns) as out:
        for num in range(100):
            sample = {'num': num}
            out.write(sample)
    StreamingDataset(local=local, allow_unsafe_types=True, batch_size=1)


@pytest.mark.usefixtures('local_remote_dir')
def test_dont_allow_unsafe_types_safe(local_remote_dir: tuple[str, str]):
    local, _ = local_remote_dir
    columns = {'num': 'int'}
    with MDSWriter(out=local, columns=columns) as out:
        for num in range(100):
            sample = {'num': num}
            out.write(sample)
    StreamingDataset(local=local, batch_size=1)


@pytest.mark.usefixtures('local_remote_dir')
def test_dont_allow_unsafe_types_unsafe(local_remote_dir: tuple[str, str]):
    local, _ = local_remote_dir
    columns = {'num': 'pkl'}
    with MDSWriter(out=local, columns=columns) as out:
        for num in range(100):
            sample = {'num': num}
            out.write(sample)
    with pytest.raises(ValueError, match='.*contains an unsafe type.*'):
        StreamingDataset(local=local, batch_size=1)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""ImageNet streaming dataset conversion scripts."""

import os
from argparse import ArgumentParser, Namespace
from glob import glob
from typing import Optional

import numpy as np
from PIL import Image
from tqdm import tqdm

from streaming.base import MDSWriter
from streaming.base.util import get_list_arg


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Args:
        Namespace: command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument(
        '--in_root',
        type=str,
        required=True,
        help='Local directory path of the input raw dataset',
    )
    args.add_argument(
        '--out_root',
        type=str,
        required=True,
        help='Directory path to store the output dataset',
    )
    args.add_argument(
        '--splits',
        type=str,
        default='train,val',
        help='Split to use. Default: train,val',
    )
    args.add_argument(
        '--compression',
        type=str,
        default='',
        help='Compression algorithm to use. Default: None',
    )
    args.add_argument(
        '--hashes',
        type=str,
        default='sha1,xxh64',
        help='Hashing algorithms to apply to shard files. Default: sha1,xxh64',
    )
    args.add_argument(
        '--size_limit',
        type=int,
        default=1 << 26,
        help='Shard size limit, after which point to start a new shard. Default: 1 << 26',
    )
    args.add_argument(
        '--progress_bar',
        type=int,
        default=1,
        help='tqdm progress bar. Default: 1 (True)',
    )
    args.add_argument(
        '--leave',
        type=int,
        default=0,
        help='Keeps all traces of the progressbar upon termination of iteration. Default: 0 ' +
        '(False)',
    )
    args.add_argument(
        '--validate',
        type=int,
        default=1,
        help='Validate that it is an Image. Default: 1 (True)',
    )
    args.add_argument(
        '--extensions',
        type=str,
        default='jpeg',
        help='Validate filename extensions. Default: jpeg',
    )
    return args.parse_args()


def check_extensions(filenames: list[str], extensions: set[str]) -> None:
    """Validate filename extensions.

    Args:
        filenames (List[str]): List of files.
        extensions (Set[str]): Acceptable extensions.
    """
    for f in filenames:
        idx = f.rindex('.')
        ext = f[idx + 1:]
        assert ext.lower() in extensions


def get_classes(filenames: list[str],
                class_names: Optional[list[str]] = None) -> tuple[list[int], list[str]]:
    """Get the classes for a dataset split of sample image filenames.

    Args:
        filenames (List[str]): Files, in the format ``"root/split/class/sample.jpeg"``.
        class_names (List[str], optional): List of class names from the other splits that we must
            match. Defaults to ``None``.

    Returns:
        Tuple[List[int], List[str]]: Class ID per sample, and the list of unique class names.
    """
    classes = []
    dirname2class = {}
    for f in filenames:
        d = f.split(os.path.sep)[-2]
        c = dirname2class.get(d)
        if c is None:
            c = len(dirname2class)
            dirname2class[d] = c
        classes.append(c)
    new_class_names = sorted(dirname2class)
    if class_names is not None:
        assert class_names == new_class_names
    return classes, new_class_names


def main(args: Namespace) -> None:
    """Main: create streaming ImageNet dataset.

    Args:
        args (Namespace): command-line arguments.
    """
    splits = get_list_arg(args.splits)
    columns = {'i': 'int', 'x': 'jpeg', 'y': 'int'}
    hashes = get_list_arg(args.hashes)
    extensions = set(get_list_arg(args.extensions))
    class_names = None
    for split in splits:
        pattern = os.path.join(args.in_root, split, '*', '*')
        filenames = sorted(glob(pattern))
        check_extensions(filenames, extensions)
        classes, class_names = get_classes(filenames, class_names)
        indices = np.random.permutation(len(filenames))
        if args.progress_bar:
            indices = tqdm(indices, leave=args.leave)
        out_split_dir = os.path.join(args.out_root, split)
        with MDSWriter(out=out_split_dir,
                       columns=columns,
                       compression=args.compression,
                       hashes=hashes,
                       size_limit=args.size_limit,
                       progress_bar=args.progress_bar) as out:
            for i in indices:
                if args.validate:
                    x = Image.open(filenames[i])
                x = open(filenames[i], 'rb').read()
                y = classes[i]
                out.write({
                    'i': int(i),
                    'x': x,
                    'y': y,
                })


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Information about nodes, ranks, and workers."""

from typing import Any

from torch.utils.data import get_worker_info
from typing_extensions import Self

from streaming.base import distributed as dist


class World:
    """Information about the nodes, ranks and workers of this run.

    .. warning::
      Be careful as to whether this object was initialized in a worker (if workers are used)
      or in a rank (which will claim one worker per rank).

    .. warning::
      In this World object, the counts (num_nodes, num_ranks, num_workers) are global -- not
      to be confused with DataLoader num_workers, which is per rank.

    Nodes are all assumed to contain the same number of devices (via local_world_size).

    Nodes:
      - node / num_nodes
      - is_multinode

    Ranks:
      - rank / num_ranks
      - rank_of_node / ranks_per_node

    Workers:
      - worker / num_workers
      - worker_of_node / workers_per_node
      - worker_of_rank / workers_per_rank
      - is_leader
      - is_local_leader
    """

    def __init__(
        self,
        num_nodes: int,
        ranks_per_node: int,
        workers_per_rank: int,
        worker: int,
    ) -> None:
        self.node = worker // (ranks_per_node * workers_per_rank)
        self.num_nodes = num_nodes
        self.is_multinode = 1 < num_nodes

        self.rank = worker // workers_per_rank
        self.num_ranks = num_nodes * ranks_per_node
        self.rank_of_node = self.rank % ranks_per_node
        self.ranks_per_node = ranks_per_node

        self.worker = worker
        self.num_workers = num_nodes * ranks_per_node * workers_per_rank
        self.worker_of_node = self.worker % (ranks_per_node * workers_per_rank)
        self.workers_per_node = ranks_per_node * workers_per_rank
        self.worker_of_rank = self.worker % workers_per_rank
        self.workers_per_rank = workers_per_rank
        self.is_leader = not worker
        self.is_local_leader = not self.worker_of_node

    def to_json(self) -> dict[str, Any]:
        """Get a JSON version of this config.

        Returns:
            Dict[str, Any]: JSON config.
        """
        return dict(self.__dict__)

    @classmethod
    def _get_worker_info(cls) -> tuple[int, int]:
        """Get worker info, or default to 0 of 1.

        Returns:
            Tuple[int, int]: Worker ID out of how many workers.
        """
        info = get_worker_info()
        if info:
            ret = info.id, info.num_workers
        else:
            ret = 0, 1
        return ret

    @classmethod
    def detect(cls) -> Self:
        """Detect the world state.

        Returns:
            Self: A new World state object according to dist and get_worker_info().
        """
        rank = dist.get_rank()
        ranks_per_node = dist.get_local_world_size()
        num_nodes = dist.get_world_size() // ranks_per_node
        worker_of_rank, workers_per_rank = cls._get_worker_info()
        worker = rank * workers_per_rank + worker_of_rank
        return cls(num_nodes, ranks_per_node, workers_per_rank, worker)

    def copy(self) -> Self:
        """Get a copy of this world state.

        Returns:
            Self: A new copy with the same state.
        """
        return World(
            num_nodes=self.num_nodes,
            ranks_per_node=self.ranks_per_node,
            workers_per_rank=self.workers_per_rank,
            worker=self.worker,
        )

    def replicate(self, replication: int) -> Self:
        """Get a copy of this world state with the given replication factor.

        Args:
            replication (int): Replication factor -- how many consecutive devices that should see
                the same samples..

        Returns:
            Self: A new sample replication version of this World state object.
        """
        if replication <= 0:
            raise ValueError(f'Replication factor must be positive.')

        if self.num_ranks % replication:
            raise ValueError(f'World size must be divisible by your replication factor.')

        rank = self.rank // replication  # Evenly divide ranks.
        num_ranks = self.num_ranks // replication  # Floor divide our rank.
        worker = rank * self.workers_per_rank + self.worker_of_rank  # Derive worker.
        # Attempt to evenly divide ranks per node. If not possible, the World
        # object will just use one node.
        if num_ranks % self.ranks_per_node == 0:
            num_nodes = num_ranks // self.ranks_per_node
        else:
            num_nodes = 1
        ranks_per_node = num_ranks // num_nodes  # Evenly divide ranks per node.
        return World(
            num_nodes=num_nodes,
            ranks_per_node=ranks_per_node,
            workers_per_rank=self.workers_per_rank,
            worker=worker,
        )

    def detect_workers(self) -> Self:
        """Get a copy of this world state with the worker information newly detected.

        Returns:
            Self: A new workers-newly-detected version of this World state object.
        """
        worker_of_rank, workers_per_rank = self._get_worker_info()
        worker = self.rank * workers_per_rank + worker_of_rank
        return World(
            num_nodes=self.num_nodes,
            ranks_per_node=self.ranks_per_node,
            workers_per_rank=workers_per_rank,
            worker=worker,
        )



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Invoke create_pretraining_data.py on shard ranges on all the cores."""

import os
from argparse import ArgumentParser, Namespace


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Returns:
        Namespace: Command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument('--in_root',
                      type=str,
                      required=True,
                      help='Path to a valid results4/ directory containing the input data.')
    args.add_argument('--in_pattern',
                      type=str,
                      default='part-%05d-of-00500',
                      help='Input shard basename pattern.')
    args.add_argument('--in_num_shards',
                      type=int,
                      default=500,
                      help='Number of input shards.')
    args.add_argument('--out_root',
                      type=str,
                      required=True,
                      help='Output directory containing shard dirs named like group-###.')
    args.add_argument('--out_pattern',
                      type=str,
                      default='group-%03d',
                      help='Output shard group dirname pattern.')
    args.add_argument('--compression',
                      type=str,
                      default='zstd:16',
                      help='Compression algorithm to use on output shards, if any. Default: ' +
                           'zstd:16.')
    args.add_argument('--hashes',
                      type=str,
                      default='sha1,xxh3_64',
                      help='Hash algorithms to use on output shards, if any. Default: ' +
                           'sha1,xxh3_64.')
    args.add_argument('--size_limit',
                      type=int,
                      default=1 << 27,
                      help='Shard size limit in bytes, after which point a new shard is ' +
                           'started. Default: 1 << 27.')
    args.add_argument('--vocab_file',
                      type=str,
                      default='vocab.txt',
                      help='The vocabulary file that the BERT model was trained on.')
    args.add_argument('--do_lower_case',
                      type=int,
                      default=1,
                      help='Whether to lower case the input text. Should be True for uncased ' +
                           'models and False for cased models.')
    args.add_argument('--max_seq_length',
                      type=int,
                      default=512,
                      help='Maximum sequence length.')
    args.add_argument('--max_predictions_per_seq',
                      type=int,
                      default=20,
                      help='Maximum number of masked LM predictions per sequence.')
    args.add_argument('--random_seed',
                      type=int,
                      default=12345,
                      help='Random seed for data generation.')
    args.add_argument('--dupe_factor',
                      type=int,
                      default=10,
                      help='Number of times to duplicate the input data (with different masks).')
    args.add_argument('--masked_lm_prob',
                      type=float,
                      default=0.15,
                      help='Masked LM probability.')
    args.add_argument('--short_seq_prob',
                      type=float,
                      default=0.1,
                      help='Probability of creating sequences which are shorter than the ' +
                           'maximum length.')
    return args.parse_args()


def main(args: Namespace) -> None:
    """Invoke create_pretraining_data.py on shard ranges in parallel.

    Args:
        args (Namespace): Command-line arguments.
    """
    num_cpus = os.cpu_count() or 1
    for cpu in range(num_cpus):
        begin = cpu * args.in_num_shards // num_cpus
        end = (cpu + 1) * args.in_num_shards // num_cpus
        input_files = ','.join([os.path.join(args.in_root, args.in_pattern % i)
                                for i in range(begin, end)])
        output_dir = os.path.join(args.out_root, args.out_pattern % cpu)
        command = f'''
            python3 create_pretraining_data.py \
                --input_file {input_files} \
                --output_dir {output_dir} \
                --compression {args.compression} \
                --hashes {args.hashes} \
                --size_limit {args.size_limit} \
                --vocab_file {args.vocab_file} \
                --do_lower_case {'True' if args.do_lower_case else 'False'} \
                --max_seq_length {args.max_seq_length} \
                --max_predictions_per_seq {args.max_predictions_per_seq} \
                --random_seed {args.random_seed} \
                --dupe_factor {args.dupe_factor} \
                --masked_lm_prob {args.masked_lm_prob} \
                --short_seq_prob {args.short_seq_prob} &
        '''
        assert not os.system(command)


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

""":class:`JSONWriter` writes samples to `.json` files that can be read by :class:`JSONReader`."""

import json
from typing import Any, Optional, Union

import numpy as np

from streaming.base.format.base.writer import SplitWriter
from streaming.base.format.json.encodings import is_json_encoded, is_json_encoding

__all__ = ['JSONWriter']


class JSONWriter(SplitWriter):
    r"""Writes a streaming JSON dataset.

    Args:
        columns (Dict[str, str]): Sample columns.
        newline (str): Newline character inserted between samples. Defaults to ``\\n``.
        out (str | Tuple[str, str]): Output dataset directory to save shard files.

            1. If ``out`` is a local directory, shard files are saved locally.
            2. If ``out`` is a remote directory, a local temporary directory is created to
               cache the shard files and then the shard files are uploaded to a remote
               location. At the end, the temp directory is deleted once shards are uploaded.
            3. If ``out`` is a tuple of ``(local_dir, remote_dir)``, shard files are saved in the
               `local_dir` and also uploaded to a remote location.
        keep_local (bool): If the dataset is uploaded, whether to keep the local dataset directory
            or remove it after uploading. Defaults to ``False``.
        compression (str, optional): Optional compression or compression:level. Defaults to
            ``None``.
        hashes (List[str], optional): Optional list of hash algorithms to apply to shard files.
            Defaults to ``None``.
        size_limit (Union[int, str], optional): Optional shard size limit, after which point to
            start a new shard. If None, puts everything in one shard. Can specify bytes
            human-readable format as well, for example ``"100kb"`` for 100 kilobyte
            (100*1024) and so on. Defaults to ``1 << 26``.
        **kwargs (Any): Additional settings for the Writer.

            progress_bar (bool): Display TQDM progress bars for uploading output dataset files to
                a remote location. Default to ``False``.
            max_workers (int): Maximum number of threads used to upload output dataset files in
                parallel to a remote location. One thread is responsible for uploading one shard
                file to a remote location. Default to ``min(32, (os.cpu_count() or 1) + 4)``.
            exist_ok (bool): If the local directory exists and is not empty, whether to overwrite
                the content or raise an error. `False` raises an error. `True` deletes the
                content and starts fresh. Defaults to `False`.
    """

    format = 'json'

    def __init__(self,
                 *,
                 columns: dict[str, str],
                 newline: str = '\n',
                 out: Union[str, tuple[str, str]],
                 keep_local: bool = False,
                 compression: Optional[str] = None,
                 hashes: Optional[list[str]] = None,
                 size_limit: Optional[Union[int, str]] = 1 << 26,
                 **kwargs: Any) -> None:
        super().__init__(out=out,
                         keep_local=keep_local,
                         compression=compression,
                         hashes=hashes,
                         size_limit=size_limit,
                         **kwargs)
        for encoding in columns.values():
            assert is_json_encoding(encoding)

        self.columns = columns
        self.newline = newline

    def encode_sample(self, sample: dict[str, Any]) -> bytes:
        """Encode a sample dict to bytes.

        Args:
            sample (Dict[str, Any]): Sample dict.

        Returns:
            bytes: Sample encoded as bytes.
        """
        obj = {}
        for key, encoding in self.columns.items():
            value = sample[key]
            assert is_json_encoded(encoding, value)
            obj[key] = value
        text = json.dumps(obj, sort_keys=True) + self.newline
        return text.encode('utf-8')

    def get_config(self) -> dict[str, Any]:
        """Get object describing shard-writing configuration.

        Returns:
            Dict[str, Any]: JSON object.
        """
        obj = super().get_config()
        obj.update({'columns': self.columns, 'newline': self.newline})
        return obj

    def encode_split_shard(self) -> tuple[bytes, bytes]:
        """Encode a split shard out of the cached samples (data, meta files).

        Returns:
            Tuple[bytes, bytes]: Data file, meta file.
        """
        data = b''.join(self.new_samples)

        num_samples = np.uint32(len(self.new_samples))
        sizes = list(map(len, self.new_samples))
        offsets = np.array([0] + sizes).cumsum().astype(np.uint32)
        obj = self.get_config()
        text = json.dumps(obj, sort_keys=True)
        meta = num_samples.tobytes() + offsets.tobytes() + text.encode('utf-8')

        return data, meta



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Helper methods to get the distributed attributes."""

import os
from typing import TypeVar, cast

import torch.distributed as dist

TObj = TypeVar('TObj')

import torch
from torch import Tensor
from torch import distributed as dist

__all__ = [
    'all_gather', 'barrier', 'broadcast', 'get_rank', 'get_local_rank', 'get_local_world_size',
    'get_world_size'
]


def get_rank() -> int:
    """Returns the rank of the current process, which is on ``[0; WORLD_SIZE - 1]``.

    Returns:
        int: The rank.
    """
    return int(os.environ.get('RANK', 0))


def get_world_size() -> int:
    """Returns the world size, which is the number of processes participating in this training run.

    Returns:
        int: The world size.
    """
    return int(os.environ.get('WORLD_SIZE', 1))


def get_local_rank() -> int:
    """Returns the local rank for the current process, which is on ``[0; LOCAL_WORLD_SIZE - 1]``.

    Returns:
        int: The local rank.
    """
    return int(os.environ.get('LOCAL_RANK', 0))


def get_local_world_size() -> int:
    """Returns the local world size, which is the number of processes for the current node.

    Returns:
        int: The local world size.
    """
    return int(os.environ.get('LOCAL_WORLD_SIZE', 1))


def barrier() -> None:
    """Synchronizes all processes."""
    if dist.is_available() and dist.is_initialized():
        dist.barrier()


def broadcast(tensor: Tensor, src: int) -> None:
    """Broadcasts the tensor to the whole group.

    Args:
        tensor (Tensor): Data to be sent if src is the rank of current process, and tensor to be
            used to save received data otherwise.
        src (int): Source rank.
    """
    if dist.is_available() and dist.is_initialized():
        dist.broadcast(tensor, src)


def all_gather(tensor_list: list[Tensor], tensor: Tensor) -> None:
    """Gathers tensors from the whole group in a list.

    Args:
        tensor_list (list[Tensor]): Output list. It should contain correctly-sized tensors to be
            used for output of the collective.
        tensor (Tensor): Tensor to be broadcast from current process.
    """
    if dist.is_available() and dist.is_initialized():
        dist.all_gather(tensor_list, tensor)


def all_gather_object(obj: TObj) -> list[TObj]:
    """Collect a pickle-able object from each rank and return a list of these objects.

    .. seealso:: :func:`torch.distributed.all_gather_object`

    Args:
        obj (TObj): Object to be gathered.

    Returns:
        List[TObj]: A list of objects indexed by rank.
    """
    if dist.is_available() and dist.is_initialized():
        obj_gather_list = [0 for _ in range(get_world_size())]
        dist.all_gather_object(obj_gather_list, obj)
        # torch.distributed will replace the None's in obj_gather_list with the gathered objects on
        # rank zero or will just be None on non-rank-zero.
        return cast(list[TObj], obj_gather_list)
    world_size = get_world_size()
    if world_size == 1:
        return [obj]
    raise RuntimeError(''.join([
        f'The world_size({world_size}) > 1, but the distributed package is not available ',
        'or has not been initialized. Please check you have initialized the distributed ',
        'runtime and that PyTorch has been built with distributed support.'
    ]))


def maybe_init_dist() -> bool:
    """Initialize torch.distributed ourselves, if necessary.

    Returns:
        bool: Whether we initialized dist ourselves.
    """
    if get_world_size() == 1 or not dist.is_available() or dist.is_initialized():
        return False
    if torch.cuda.is_available() and dist.is_nccl_available():
        backend = 'nccl'
    else:
        backend = 'gloo'
    dist.init_process_group(backend=backend, rank=get_rank(), world_size=get_world_size())
    return True



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""sim_cli: simulate your training yaml from the command line."""

import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import argparse

import humanize
from core.main import simulate
from core.utils import get_simulation_stats
from core.yaml_processing import create_simulation_dataset, ingest_yaml
from interfaces.interface_utils import plot_simulation

from streaming.base.util import bytes_to_int

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Simulate your training yaml from the command \
                                     line.')
    parser.add_argument('-f', '--file', type=str, help='path to yaml file', required=True)
    parser.add_argument('-n', '--nodes', type=int, help='number of physical nodes', required=False)
    parser.add_argument('-d',
                        '--devices',
                        type=int,
                        help='number of devices per node',
                        required=False)
    parser.add_argument('-t',
                        '--time_per_sample',
                        type=float,
                        help='time to process one sample on one device (seconds)',
                        required=False)
    parser.add_argument('-b',
                        '--node_internet_bandwidth',
                        type=str,
                        help='internet bandwidth per node (bytes/s)',
                        required=False)
    args = parser.parse_args()

    # Read in yaml file
    filepath = args.file
    total_devices, workers, max_duration, global_batch_size, train_dataset = \
        ingest_yaml(filepath=filepath)

    # Check if we have to ask for any parameters
    args = parser.parse_args()
    nodes = args.nodes
    if nodes is None:
        nodes = int(input('Number of physical nodes: '))
    # devices may be specified in the yaml file.
    if total_devices is None:
        devices = args.devices
    else:
        if total_devices % nodes != 0:
            raise ValueError('The number of devices must be divisible by the number of nodes.')
        devices = total_devices // nodes
    time_per_sample = args.time_per_sample
    node_network_bandwidth = args.node_internet_bandwidth
    if devices is None:
        devices = int(input('Number of devices per node: '))
    if time_per_sample is None:
        time_per_sample = float(input('Time to process one sample on one device (seconds): '))
    if node_network_bandwidth is None:
        bandwidth_input = input('Internet bandwidth per node (bytes/s): ')
        try:
            # Converting to float first handles the case where the input is a string in scientific
            # notation, like "1e9".
            node_network_bandwidth = float(bandwidth_input)
            node_network_bandwidth = int(node_network_bandwidth)
        except ValueError:
            node_network_bandwidth = str(bandwidth_input)

    # Convert strings into numbers for applicable args
    node_network_bandwidth = bytes_to_int(node_network_bandwidth)

    # Create SimulationDataset
    print('Constructing SimulationDataset...')
    dataset = create_simulation_dataset(nodes, devices, workers, global_batch_size, train_dataset)

    # Simulate Run
    results = next(simulate(dataset, time_per_sample, node_network_bandwidth, max_duration))
    if len(results) != 4:
        raise ValueError(f'Simulation with generate=False should return 4 final results. ' +
                         f'Instead, received `results` of length {len(results)}.')
    step_times, step_downloads, startup_time, min_cache_limit = results

    print('Simulation Finished.')

    # Display simulation stats
    total_batches = len(step_times)
    cache_limit = dataset.get_cache_limit()
    all_throughput_drops, warmup_time, warmup_step, post_warmup_throughput_drops = \
        get_simulation_stats(step_times, time_per_sample, global_batch_size//(nodes*devices))
    print('\nSimulation Stats:')
    print(f'Minimum cache limit needed: {humanize.naturalsize(min_cache_limit)}')
    if cache_limit is not None and cache_limit < min_cache_limit:
        # Cache limit is too low, and will cause shard redownloads / throughput drops.
        print('⚠️ The provided cache limit is lower than the minimum cache limit needed to \
            prevent shard re-downloads. This can cause throughput issues.')
    if warmup_step == total_batches:
        # display error if the warmup phase is the whole run, so we never hit peak throughput.
        print('🚨 This configuration is severely bottlenecked by downloading. The run will not be \
              performant.')
    elif post_warmup_throughput_drops:
        # display warning if post-warmup throughput drops are more than 10% of the run.
        print('⚠️ This configuration experiences some downloading-related slowdowns even after \
              warmup.')
    print('{0} steps, or {1:.1f}% of all steps, waited for shard downloads.'\
          .format(all_throughput_drops, 100 * all_throughput_drops / (total_batches)))
    if warmup_step != total_batches:
        # only display post-warmup throughput drop info if we actually ended the warmup period
        # (i.e. we hit peak throughput at some point)
        print('There were {} steps that waited for shard downloads after the warmup period.'\
              .format(post_warmup_throughput_drops))
    print('Estimated time to first batch: {0:.2f} s'.format(startup_time))
    print('Estimated warmup time: {0:.2f} s'.format(warmup_time))

    # Plot simulation
    plot_simulation(step_times, step_downloads)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import pathlib
import time
from typing import Any, Optional

import pytest

from streaming.base import StreamingDataset
from streaming.text import StreamingC4
from streaming.vision import StreamingADE20K, StreamingCIFAR10, StreamingCOCO, StreamingImageNet


def get_dataset(name: str,
                local: str,
                split: str,
                shuffle: bool,
                batch_size: Optional[int],
                other_kwargs: Optional[dict[str, Any]] = None) -> tuple[int, StreamingDataset]:
    other_kwargs = {} if other_kwargs is None else other_kwargs
    dataset_map = {
        'ade20k': {
            'remote': 's3://mosaicml-internal-dataset-ade20k/mds/2/',
            'num_samples': {
                'train': 20206,
                'val': 2000,
            },
            'class': StreamingADE20K,
            'kwargs': {},
        },
        'imagenet1k': {
            'remote': 's3://mosaicml-internal-dataset-imagenet1k/mds/2/',
            'num_samples': {
                'train': 1281167,
                'val': 50000,
            },
            'class': StreamingImageNet,
            'kwargs': {},
        },
        'coco': {
            'remote': 's3://mosaicml-internal-dataset-coco/mds/2/',
            'num_samples': {
                'train': 117266,
                'val': 4952,
            },
            'class': StreamingCOCO,
            'kwargs': {},
        },
        'c4': {
            'remote': 's3://mosaicml-internal-dataset-c4/mds/2/',
            'num_samples': {
                'train': 364868892,
                'val': 364608,
            },
            'class': StreamingC4,
            'kwargs': {
                'tokenizer_name': 'bert-base-uncased',
                'max_seq_len': 512,
                'group_method': 'truncate'
            },
        },
        'cifar10': {
            'remote': 's3://mosaicml-internal-dataset-cifar10/mds/2/',
            'num_samples': {
                'train': 50000,
                'val': 10000,
            },
            'class': StreamingCIFAR10,
            'kwargs': {},
        },
        'test_streaming_upload': {
            'remote': 's3://streaming-upload-test-bucket/',
            'num_samples': {
                'all': 0,
            },
            'class': StreamingDataset,
            'kwargs': {},
        }
    }
    if name not in dataset_map and split not in dataset_map[name]['num_samples'][split]:
        raise ValueError('Could not load dataset with name={name} and split={split}')

    d = dataset_map[name]
    expected_samples = d['num_samples'][split]
    remote = d['remote']
    kwargs = {**d['kwargs'], **other_kwargs}
    dataset = d['class'](local=local,
                         remote=remote,
                         split=split,
                         shuffle=shuffle,
                         batch_size=batch_size,
                         **kwargs)
    return (expected_samples, dataset)


@pytest.mark.remote
@pytest.mark.parametrize('name', [
    'ade20k',
    'imagenet1k',
    'coco',
    'cifar10',
    'c4',
])
@pytest.mark.parametrize('split', ['val'])
def test_streaming_remote_dataset(tmp_path: pathlib.Path, name: str, split: str) -> None:
    # Build StreamingDataset
    build_start = time.time()
    expected_samples, dataset = get_dataset(name=name,
                                            local=str(tmp_path),
                                            split=split,
                                            shuffle=False,
                                            batch_size=None)
    build_end = time.time()
    build_dur = build_end - build_start
    print('Built dataset')

    # Test basic iteration
    rcvd_samples = 0
    iter_start = time.time()
    for _ in dataset:
        rcvd_samples += 1

        if (rcvd_samples % 1000 == 0):
            print(f'samples read: {rcvd_samples}')

    iter_end = time.time()
    iter_dur = iter_end - iter_start
    samples_per_sec = rcvd_samples / iter_dur

    # Print debug info
    print(f'build_dur={build_dur:.2f}s, iter_dur={iter_dur:.2f}, ' +
          f'samples_per_sec={samples_per_sec:.2f}')

    # Test all samples arrived
    assert rcvd_samples == expected_samples
